"""
This module contains a parser for, and methods to extract information from
ELF files.
"""

from pwnypack.target import Target
from pwnypack.packing import U16, U32, unpack, pack_size
import pwnypack.main
from enum import IntEnum


__all__ = [
    'ELF',
]


class ELF(Target):
    """
    A parser for ELF files. Upon parsing the ELF headers, it will not only
    fill the ELF specific fields but will also populate the inherited
    :attr:`~pwnypack.target.Target.arch`, :attr:`~pwnypack.target.Target.bits`
    and :attr:`~pwnypack.target.Target.endian` properties based on the
    values it encounters.

    Arguments:
        f(str, file or ``None``): The (path to) the ELF file to parse.

    Example:
        >>> from pwny import *
        >>> e = ELF('my-executable')
        >>> print(e.machine)
        >>> print(e.program_headers)
        >>> print(e.section_headers)
        >>> print(e.symbols)
    """

    class ProgramHeader(object):
        """
        Describes how the loader will load a part of a file. Called by the
        :class:`ELF` class.

        Args:
            elf(ELF): The ELF instance owning this program header.
            data: The content of the program header entry.
        """

        class Type(IntEnum):
            """
            The segment type.
            """

            unknown = -1               #: Unknown type, check type_id for exact type
            null = 0                   #: The element is unused
            load = 1                   #: The element contains a loadable segment
            dynamic = 2                #: The element contains dynamic linking information
            interp = 3                 #: The element contains the path of the interpreter
            note = 4                   #: The element contains auxiliary information
            shlib = 5                  #: This element type is reserved
            phdr = 6                   #: This element contains the program header table itself
            gnu_eh_frame = 0x6474e550  #: This element contains the exception handler unwind information
            gnu_stack = 0x6474e551     #: This element describes the access right of the stack
            gnu_relro = 0x6474e552     #: This element contains the readonly relocations

        class Flags(IntEnum):
            """
            The individual flags that make up :attr:`ELF.ProgramHeader.flags`.
            """

            x = 1                  #: Segment is executable
            w = 2                  #: Segment is writable
            r = 4                  #: Segment is readable

        type = None     #: The type of the segment (:class:`~ELF.ProgramHeader.Type`).
        type_id = None  #: The numerical type describing the segment.
        offset = None   #: Where in the file the segment is located.
        vaddr = None    #: The virtual address at which the segment is loaded.
        paddr = None    #: The physical address at which the segment is loaded.
        filesz = None   #: The size of the segment in the file.
        memsz = None    #: The size of the segment in memory.
        flags = None    #: The flags for the segment (OR'ed values of :class:`~ELF.ProgramHeader.Flags`).
        align = None    #: The alignment of the segment.

        def __init__(self, elf, data):
            if elf.bits == 32:
                fmt = 'IIIIIIII'
                fmt_fields = [
                    'type_id',
                    'offset',
                    'vaddr',
                    'paddr',
                    'filesz',
                    'memsz',
                    'flags',
                    'align',
                ]
            else:
                fmt = 'IIQQQQQQ'
                fmt_fields = [
                    'type_id',
                    'flags',
                    'offset',
                    'vaddr',
                    'paddr',
                    'filesz',
                    'memsz',
                    'align',
                ]
            fmt_size = pack_size(fmt)

            for key, value in zip(
                fmt_fields,
                unpack(fmt, data[:fmt_size], target=elf)
            ):
                setattr(self, key, value)

            try:
                self.type = self.Type(self.type_id)
            except ValueError:
                self.type = self.Type.unknown

    class SectionHeader(object):
        """
        Describes a section of an ELF file. Called by the :class:`ELF` class.

        Args:
            elf(ELF): The ELF instance owning this section header.
            data: The content of the section header entry.
        """

        class Type(IntEnum):
            """
            Describes the section's type
            """

            unknown = -1                  #: Unknown section type
            null = 0                      #: Inactive section header
            progbits = 1                  #: Program defined information
            symtab = 2                    #: Full symbol table
            strtab = 3                    #: String table
            rela = 4                      #: Relocation entries with explicit addends
            hash = 5                      #: Symbol hash table
            dynamic = 6                   #: Dynamic linking information
            note = 7                      #: Vendor or system specific notes
            nobits = 8                    #: Occupies no file space, initialised to 0
            rel = 9                       #: Relocation entries without explicit addends
            dynsym = 11                   #: Minimal symbol table for dynamic linking
            init_array = 14               #: Array of initialisation functions
            fini_array = 15               #: Array of termination functions
            preinit_array = 16            #: Array of initialisation functions
            group = 17                    #: Section group
            symtab_shndx = 18             #: Extended symbol section mapping table
            num = 19                      #: Number of defined types

            checksum = 0x6ffffff8         #: Checksum for DSO content

            # Sun extensions:
            sunw_move = 0x6ffffffa        #: SUN extension: Additional information for partially initialized data.
            sunw_comdat = 0x6ffffffb      #: SUN extension
            sunw_syminfo = 0x6ffffffc     #: SUN extension: Extra symbol information.

            # GNU extensions:
            gnu_attributes = 0x6ffffff5   #: GNU extension: Object attributes
            gnu_hash = 0x6ffffff6         #: GNU extension: GNU-style hash section
            gnu_liblist = 0x6ffffff7      #: GNU extension: Pre-link library list
            gnu_object_only = 0x6ffffff8  #: GNU extension
            gnu_verdef = 0x6ffffffd       #: GNU extension: Version definition section
            gnu_verneed = 0x6ffffffe      #: GNU extension: Version requirements section
            gnu_versym = 0x6fffffff       #: GNU extension: Version symbol table

        class Flags(IntEnum):
            write = 1 << 0             #: Section is writable
            alloc = 1 << 1             #: Section occupies memory during execution
            execinstr = 1 << 2         #: Section contains executable code
            merge = 1 << 4             #: Section might be merged
            strings = 1 << 5           #: Section contains NUL terminated strings
            info_link = 1 << 6         #: Section's :attr:`~ELF.SectionHeader.info` field contains SHT index
            link_order = 1 << 7        #: Preserve section order after combining
            os_nonconforming = 1 << 8  #: Non-standard OS-specific handling required
            group = 1 << 9             #: Section is member of a group
            tls = 1 << 10              #: Section holds thread-local data
            maskos = 0x0ff00000        #: Mask for OS specific flags
            maskproc = 0xf0000000      #: Mask for processor specific flags
            ordered = 1 << 30          #: Treat sh_link, sh_info specially
            exclude = 1 << 31          #: Exclude section from linking

        elf = None         #: The instance of :class:`ELF` this symbol belongs to

        name = None        #: The name of this section
        name_index = None  #: The index into the string table for this section's name
        type = None        #: The type of this section (one of :class:`~ELF.SectionHeader.Type`
        type_id = None     #: The numeric identifier of the section type
        flags = None       #: The flags for this section, see :class:`~ELF.SectionHeader.Flags`
        addr = None        #: The memory address at which this section will be loaded
        offset = None      #: The offset in the file where this section resides
        size = None        #: The size of this section in the file
        link = None        #: Holds a section type dependant section header table index link
        info = None        #: Holds section type dependant extra information
        addralign = None   #: Address alignment constraint
        entsize = None     #: Size of the entries in this section

        _content = None

        def __init__(self, elf, data):
            self.elf = elf

            if elf.bits == 32:
                fmt = 'IIIIIIIIII'
            else:
                fmt = 'IIQQQQIIQQ'
            fmt_size = pack_size(fmt)

            for key, value in zip(
                [
                    'name_index',
                    'type_id',
                    'flags',
                    'addr',
                    'offset',
                    'size',
                    'link',
                    'info',
                    'addralign',
                    'entsize',
                ],
                unpack(fmt, data[:fmt_size], target=elf)
            ):
                setattr(self, key, value)

            try:
                self.type = self.Type(self.type_id)
            except ValueError:
                self.type = self.Type.unknown

        @property
        def content(self):
            """
            The contents of this section.
            """

            if self._content is None:
                self.elf.f.seek(self.offset)
                self._content = self.elf.f.read(self.size)

            return self._content

    class Symbol(object):
        """
        Contains information about symbols. Called by the :class:`ELF` class.

        Args:
            elf(ELF): The ELF instance owning this symbol.
            data: The content of the symbol definition.
            strs: The content of the string section associated with the symbol table.
        """

        class Binding(IntEnum):
            """
            Describes a symbol's binding.
            """

            local = 0    #: Local symbol
            global_ = 1  #: Global symbol
            weak = 2     #: Weak symbol

        class Type(IntEnum):
            """
            Describes the symbol's type.
            """

            unknown = -1  #: Symbol has an unknown type
            notype = 0    #: Symbol has no type
            object = 1    #: Symbol is an object
            func = 2      #: Symbol is a function or contains other executable code
            section = 3   #: Symbol is associated with a section
            file = 4      #: Contains the name of the source file
            common = 5    #: The symbol labels an uninitialized common block
            tls = 6       #: The symbol specifies a Thread-Local Storage entity

        class Visibility(IntEnum):
            """
            Describes the symbol's visibility.
            """

            default = 0    #: Global and weak symbols are visible, local symbols are hidden
            internal = 1   #: Symbol is an internal symbol
            hidden = 2     #: Symbol is invisible to other components
            protected = 3  #: Symbol is visible but not preemptable

        class SpecialSection(IntEnum):
            """
            Special section types.
            """

            undef = 0        #: Symbol is undefined and will be resolved by the runtime linker
            abs = 0xfff1     #: Symbol has an absolute value that will not change because of relocation
            common = 0xfff2  #: Symbol labels a common block that has not yet been allocated.

        elf = None         #: The instance of :class:`ELF` this symbol belongs to

        name_index = None  #: The index of the symbol's name in the string table
        value = None       #: The value of the symbol (type dependent)
        size = None        #: The size of the symbol
        info = None        #: Describes the symbol's type and binding (see :attr:`~ELF.Symbol.type` and
                           #: :attr:`ELF.Symbol.binding`)
        other = None       #: Specifies the symbol's visibility
        shndx = None
        """
        The section in which this symbol is defined (or one of the :class:`~ELF.Symbol.SpecialSection` types)
        """

        name = None        #: The resolved name of this symbol
        type_id = None     #: The numerical type of this symbol
        type = None        #: The resolved type of this symbol (one of :class:`~ELF.Symbol.Type`)
        visibility = None  #: The visibility of this symbol (one of :class:`~ELF.Symbol.Visibility`)

        _content = None

        def __init__(self, elf, data, strs):
            self.elf = elf

            if elf.bits == 32:
                fmt = 'IIIBBH'
            else:
                fmt = 'IBBHQQ'

            if elf.bits == 32:
                self.name_index, self.value, self.size, self.info, self.other, self.shndx = \
                    unpack(fmt, data, target=elf)
            else:
                self.name_index, self.info, self.other, self.shndx, self.value, self.size = \
                    unpack(fmt, data, target=elf)

            self.type_id = self.info & 15
            self.binding = self.Binding(self.info >> 4)
            self.visibility = self.Visibility(self.other & 3)

            self.name = strs[self.name_index:].split('\0', 1)[0]

            try:
                self.type = self.Type(self.type_id)
            except ValueError:
                self.type = self.Type.unknown

        @property
        def content(self):
            """
            The contents of a symbol.

            Raises:
                TypeError: If the symbol isn't defined until runtime.
            """

            if self.shndx in (self.SpecialSection.undef, self.SpecialSection.abs,
                              self.SpecialSection.common):
                raise TypeError('Symbol is not defined')

            if self._content is None:
                section_header = self.elf.get_section_header(self.shndx)
                symbol_offset = self.value - section_header.addr

                self.elf.f.seek(section_header.offset + symbol_offset)
                self._content = self.elf.f.read(self.size)

            return self._content

    class Type(IntEnum):
        """
        Describes the object type.
        """

        unknown = -1     #: Unknown object type
        none = 0         #: No file type
        relocatable = 1  #: Relocatable file
        executable = 2   #: Executable file
        shared = 3       #: Shared object file
        core = 4         #: Core file
        os = 0xfe00      #: OS specific
        proc = 0xff00    #: Processor specific

    class Machine(IntEnum):
        """
        The target machine architecture.
        """

        unknown = -1      #: Unknown architecture
        none = 0          #: No machine
        m32 = 1           #: AT&T WE 32100
        sparc = 2         #: SPARC
        i386 = 3          #: Intel 80386
        m68k = 4          #: Motorola 68000
        m88k = 5          #: Motorola 88000
        i860 = 7          #: Intel 80860
        mips = 8          #: MIPS I Architecture
        s370 = 9          #: IBM System/370 Processor
        mips_rs3_le = 10  #: MIPS RS3000 Little-endian
        parisc = 15       #: Hewlett-Packard PA-RISC
        vpp550 = 17       #: Fujitsu VPP500
        sparc32plus = 18  #: Enhanced instruction set SPARC
        i960 = 19         #: Intel 80960
        ppc = 20          #: PowerPC
        ppc64 = 21        #: 64-bit PowerPC
        s390 = 22         #: IBM System/390 Processor
        v800 = 36         #: NEC V800
        fr20 = 37         #: Fujitsu FR20
        rh32 = 38         #: TRW RH-32
        rce = 39          #: Motorola RCE
        arm = 40          #: Advanced RISC Machines ARM
        alpha = 41        #: Digital Alpha
        superh = 42       #: Hitachi SuperH
        sparcv9 = 43      #: SPARC Version 9
        tricore = 44      #: Siemens TriCore embedded processor
        arc = 45          #: Argonaut RISC Core, Argonaut Technologies Inc.
        h8_300 = 46       #: Hitachi H8/300
        h8_300h = 47      #: Hitachi H8/300H
        h8s = 48          #: Hitachi H8S
        h8_500 = 49       #: Hitachi H8/500
        ia64 = 50         #: Intel IA-64 processor architecture
        mipsx = 51        #: Stanford MIPS-X
        coldfire = 52     #: Motorola ColdFire
        m68hc12 = 53      #: Motorola M68HC12
        mma = 54          #: Fujitsu MMA Multimedia Accelerator
        pcp = 55          #: Siemens PCP
        ncpu = 56         #: Sony nCPU embedded RISC processor
        ndr1 = 57         #: Denso NDR1 microprocessor
        starcore = 58     #: Motorola Star*Core processor
        me16 = 59         #: Toyota ME16 processor
        st100 = 60        #: STMicroelectronics ST100 processor
        tinyj = 61        #: Advanced Logic Corp. TinyJ embedded processor family
        x86_64 = 62       #: AMD x86-64 architecture
        pdsp = 63         #: Sony DSP Processor
        pdp10 = 64        #: Digital Equipment Corp. PDP-10
        pdp11 = 65        #: Digital Equipment Corp. PDP-11
        fx66 = 66         #: Siemens FX66 microcontroller
        st9plus = 67      #: STMicroelectronics ST9+ 8/16 bit microcontroller
        st7 = 68          #: STMicroelectronics ST7 8-bit microcontroller
        m68hc16 = 69      #: Motorola MC68HC16 Microcontroller
        m68hc11 = 70      #: Motorola MC68HC11 Microcontroller
        m68hc08 = 71      #: Motorola MC68HC08 Microcontroller
        m68hc05 = 72      #: Motorola MC68HC05 Microcontroller
        svx = 73          #: Silicon Graphics SVx
        st19 = 74         #: STMicroelectronics ST19 8-bit microcontroller
        vax = 75          #: Digital VAX
        cris = 76         #: Axis Communications 32-bit embedded processor
        javelin = 77      #: Infineon Technologies 32-bit embedded processor
        firepath = 78     #: Element 14 64-bit DSP Processor
        zsp = 79          #: LSI Logic 16-bit DSP Processor
        mmix = 80         #: Donald Knuth's educational 64-bit processor
        huany = 81        #: Harvard University machine-independent object files
        prism = 82        #: SiTera Prism
        avr = 83          #: Atmel AVR 8-bit microcontroller
        fr30 = 84         #: Fujitsu FR30
        d10v = 85         #: Mitsubishi D10V
        d30v = 86         #: Mitsubishi D30V
        v850 = 87         #: NEC v850
        m32r = 88         #: Mitsubishi M32R
        mn10300 = 89      #: Matsushita MN10300
        mn10200 = 90      #: Matsushita MN10200
        pj = 91           #: picoJava
        openrisc = 92     #: OpenRISC 32-bit embedded processor
        arc_a5 = 93       #: ARC Cores Tangent-A5
        xtensa = 94       #: Tensilica Xtensa Architecture
        videocore = 95    #: Alphamosaic VideoCore processor
        tmm_gpp = 96      #: Thompson Multimedia General Purpose Processor
        ns32k = 97        #: National Semiconductor 32000 series
        tpc = 98          #: Tenor Network TPC processor
        snp1k = 99        #: Trebia SNP 1000 processor
        st200 = 100       #: STMicroelectronics ST200 microcontroller
        ip2k = 101        #: Ubicom IP2xxx microcontroller family
        max = 102         #: MAX Processor
        cr = 103          #: National Semiconductor CompactRISC microprocessor
        f2mc16 = 104      #: Fujitsu F2MC16
        msp430 = 105      #: Texas Instruments embedded microcontroller msp430
        blackfin = 106    #: Analog Devices Blackfin (DSP) processor
        se_c33 = 107      #: S1C33 Family of Seiko Epson processors
        sep = 108         #: Sharp embedded microprocessor
        arca = 109        #: Arca RISC Microprocessor
        unicore = 110     #: Microprocessor series from PKU-Unity Ltd. and MPRC of Peking University
        aarch64 = 183     #: 64-bit Advanced RISC Machines ARM

    class OSABI(IntEnum):
        """
        Describes the OS- or ABI-specific ELF extensions used by this file.
        """

        unknown = -1  #: Unknown ABI
        system_v = 0  #: SystemV ABI / No extensions
        hp_ux = 1     #: HP-UX ABI
        netbsd = 2    #: NetBSD ABI
        linux = 3     #: Linux ABI
        solaris = 6   #: Solaris ABI
        aix = 7       #: AIX ABI
        irix = 8      #: IRIX ABI
        freebsd = 9   #: FreeBSD ABI
        tru64 = 10    #: Compaq TRU64 Unix
        modesto = 11  #: Novell Modesto
        openbsd = 12  #: OpenBSD ABI
        openvms = 13  #: OpenVMS ABI
        nsk = 14      #: Hewlett-Packard Non-Stop Kernel
        aros = 15     #: Amiga Research OS
        arch = 64     #: Architecture specific ABI

    _ELF_MAGIC = b'\x7fELF'

    f = None  #: The ELF file.

    osabi = None        #: The OSABI (one of :class:`ELF.OSABI`).
    abi_version = None  #: The specific ABI version of the OS / ABI.
    type = None         #: The object type (one of :class:`ELF.Type`).
    machine = None      #: The machine architecture (one of :class:`ELF.Machine`).
    entry = None        #: The entry point address.
    phoff = None        #: The offset of the first program header in the file.
    shoff = None        #: The offset of the first section header in the file.
    flags = None        #: The flags. Currently, no flags are defined.
    hsize = None        #: The size of the header.
    phentsize = None    #: The size of a program header.
    phnum = None        #: The number of program headers.
    shentsize = None    #: The size of a section header.
    shnum = None        #: The number of section headers.
    shstrndx = None     #: The index of the section containing the section names.

    _section_headers_by_name = None
    _section_headers_by_index = None
    _program_headers = None
    _symbols_by_index = None
    _symbols_by_name = None

    def __init__(self, f=None):
        super(ELF, self).__init__()
        if f is not None:
            self.parse_file(f)

    def _parse_header(self, data):
        """
        Parse the ELF header in ``data`` and populate the properties.

        Args:
            data(bytes): The ELF header.
        """

        (magic, word_size, byte_order, version, osabi, abi_version, _), data = \
            unpack('4sBBBBB7s', data[:16]), data[16:]

        assert magic == self._ELF_MAGIC, 'Missing ELF magic'
        assert word_size in (1, 2), 'Invalid word size'
        assert byte_order in (1, 2), 'Invalid byte order'

        assert version == 1, 'Invalid version'

        self.osabi = self.OSABI(osabi)
        self.abi_version = abi_version

        (type_, machine, version), data = unpack('HHI', data[:8], endian=self.endian), data[8:]

        self.type = self.Type(type_)
        self.machine = ELF.Machine(machine)
        assert version == 1, 'Invalid version'

        if self.machine is ELF.Machine.i386:
            arch = Target.Arch.x86
            assert word_size == 1, 'Unexpected ELF64 for machine type x86'
            assert byte_order == 1, 'Unexpected big-endian for machine type x86'
        elif self.machine is ELF.Machine.x86_64:
            arch = Target.Arch.x86
            assert word_size == 2, 'Unexpected ELF32 for machine type x64_64'
            assert byte_order == 1, 'Unexpected big-endian for machine type x86'
        elif self.machine is ELF.Machine.arm:
            arch = Target.Arch.arm
            assert word_size == 1, 'Unexpected ELF64 for machine type arm'
        elif self.machine is ELF.Machine.aarch64:
            arch = Target.Arch.arm
            assert word_size == 2, 'Unexpected ELF32 for machine type aarch64'
        else:
            arch = Target.Arch.unknown

        self.arch = arch
        self.bits = 32 * word_size
        self.endian = byte_order - 1

        if self.bits == 32:
            fmt = 'IIIIHHHHHH'
        else:
            fmt = 'QQQIHHHHHH'

        fmt_size = pack_size(fmt)
        (self.entry, self.phoff, self.shoff, self.flags, self.hsize, self.phentsize,
            self.phnum, self.shentsize, self.shnum, self.shstrndx) = \
            unpack(fmt, data[:fmt_size], target=self)

    def parse_file(self, f):
        """
        Parse an ELF file and fill the class' properties.

        Arguments:
            f(file or str): The (path to) the ELF file to read.
        """

        if type(f) is str:
            self.f = open(f, 'rb')
        else:
            self.f = f
        self._parse_header(self.f.read(64))

    def _ensure_program_headers_loaded(self):
        if self._program_headers is not None:
            return

        self._program_headers = []

        if self.phnum:
            self.f.seek(self.phoff)
            for i in range(self.phnum):
                program_header = self.ProgramHeader(self, self.f.read(self.phentsize))
                self._program_headers.append(program_header)

    @property
    def program_headers(self):
        """
        A list of all program headers.
        """

        self._ensure_program_headers_loaded()
        return self._program_headers

    def get_program_header(self, index):
        """
        Return a specific program header by its index.

        Args:
            index(int): The program header index.

        Returns:
            :class:`~ELF.ProgramHeader`: The program header.

        Raises:
            KeyError: The specified index does not exist.
        """

        self._ensure_section_headers_loaded()
        return self._program_headers[index]

    def _ensure_section_headers_loaded(self):
        if self._section_headers_by_index is not None:
            return

        self._section_headers_by_index = []
        self._section_headers_by_name = {}

        if self.shnum:
            self.f.seek(self.shoff)
            for i in range(self.shnum):
                section_header = self.SectionHeader(self, self.f.read(self.shentsize))
                self._section_headers_by_index.append(section_header)

            strings_section = self._section_headers_by_index[self.shstrndx]
            section_strings = strings_section.content.decode('ascii')
            for section_header in self._section_headers_by_index:
                name_index = section_header.name_index
                section_header.name = name = section_strings[name_index:].split('\0', 1)[0]
                self._section_headers_by_name[name] = section_header

    @property
    def section_headers(self):
        """
        Return the list of section headers.
        """

        self._ensure_section_headers_loaded()
        return self._section_headers_by_index

    def get_section_header(self, section):
        """
        Get a specific section header by index or name.

        Args:
            section(int or str): The index or name of the section header to return.

        Returns:
            :class:`~ELF.SectionHeader`: The section header.

        Raises:
            KeyError: The requested section header does not exist.
        """

        self._ensure_section_headers_loaded()
        if type(section) is int:
            return self._section_headers_by_index[section]
        else:
            return self._section_headers_by_name[section]

    def _parse_symbols(self, syms, strs):
        symbols = []
        if self.bits == 32:
            fmt = 'IIIBBH'
        else:
            fmt = 'IBBHQQ'
        fmt_size = pack_size(fmt)

        while syms:
            sym, syms = syms[:fmt_size], syms[fmt_size:]
            symbols.append(self.Symbol(self, sym, strs))

        return symbols

    def _read_symbols(self, symbol_section, string_section=None):
        if string_section is None:
            string_section = {
                '.symtab': '.strtab',
                '.dynsym': '.dynstr'
            }.get(symbol_section, None)
            if string_section is None:
                raise ValueError('Could not determine string section for symbol section %s' % symbol_section)

        return self._parse_symbols(
            self.get_section_header(symbol_section).content,
            self.get_section_header(string_section).content.decode('ascii'),
        )

    def _ensure_symbols_loaded(self):
        if self._symbols_by_index is None:
            try:
                symbols = self._read_symbols('.symtab')
            except KeyError:
                try:
                    symbols = self._read_symbols('.dynsym')
                except KeyError:
                    symbols = []

            self._symbols_by_index = symbols
            self._symbols_by_name = dict(
                (symbol.name, symbol)
                for symbol in symbols
                if symbol.name
            )

    @property
    def symbols(self):
        """
        Return a list of all symbols.
        """

        self._ensure_symbols_loaded()
        return self._symbols_by_index

    def get_symbol(self, symbol):
        """
        Get a specific symbol by index or name.

        Args:
            symbol(int or str): The index or name of the symbol to return.

        Returns:
            ELF.Symbol: The symbol.

        Raises:
            KeyError: The requested symbol does not exist.
        """

        self._ensure_symbols_loaded()
        if type(symbol) is int:
            return self._symbols_by_index[symbol]
        else:
            return self._symbols_by_name[symbol]


@pwnypack.main.register(name='symbols')
def symbols_app(parser, _, args):  # pragma: no cover
    """
    List ELF symbol table.
    """

    parser.add_argument('file', help='ELF file to list the symbols of')
    parser.add_argument('symbol', nargs='?', help='show only this symbol')
    parser.add_argument('--exact', '-e', action='store_const', const=True, help='filter by exact symbol name')
    args = parser.parse_args(args)

    print('%-18s %5s %-7s %-7s %-10s %5s %s' % (
        'value',
        'size',
        'type',
        'binding',
        'visibility',
        'index',
        'name',
    ))

    elf = ELF(args.file)
    for symbol in elf.symbols:
        if args.symbol:
            if args.exact:
                if symbol.name != args.symbol:
                    continue
            else:
                if args.symbol.lower() not in symbol.name.lower():
                    continue

        if symbol.shndx == symbol.SpecialSection.undef:
            shndx = 'UND'
        elif symbol.shndx == symbol.SpecialSection.abs:
            shndx = 'ABS'
        elif symbol.shndx == symbol.SpecialSection.common:
            shndx = 'COM'
        else:
            shndx = str(symbol.shndx)

        print('0x%016x %5d %-7s %-7s %-10s %5s %s' % (
            symbol.value,
            symbol.size,
            symbol.type.name,
            symbol.binding.name,
            symbol.visibility.name,
            shndx,
            symbol.name,
        ))


@pwnypack.main.register(name='symbol-extract')
def extract_symbol_app(parser, _, args):  # pragma: no cover
    """
    Extract a symbol from an ELF file.
    """

    parser.add_argument('file', help='ELF file to extract a symbol from')
    parser.add_argument('symbol', help='the symbol to extract')
    args = parser.parse_args(args)
    return ELF(args.file).get_symbol(args.symbol).content



import six
import mock
import pwny


headers = [
    {
        'data': b'\x7fELF\x02\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00>\x00\x01\x00\x00\x00@\x04@\x00\x00'
                b'\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00p\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00@\x008\x00\t'
                b'\x00@\x00\x1e\x00\x1b\x00',
        'header': {
            'arch': pwny.Target.Arch.x86,
            'bits': 64,
            'endian': pwny.Target.Endian.little,
            'abi_version': 0,
            'entry': 4195392,
            'flags': 0,
            'hsize': 64,
            'osabi': pwny.ELF.OSABI.system_v,
            'phentsize': 56,
            'phnum': 9,
            'phoff': 64,
            'shentsize': 64,
            'shnum': 30,
            'shoff': 4464,
            'shstrndx': 27,
            'type': pwny.ELF.Type.executable
        },
    },
    {
        'data': b'\x7fELF\x02\x01\x01\x03\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00>\x00\x01\x00\x00\x00N\x0f@\x00\x00'
                b'\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00\xa8\x1d\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00@\x008\x00'
                b'\x06\x00@\x00\x1f\x00\x1c\x00',
        'header': {
            'arch': pwny.Target.Arch.x86,
            'bits': 64,
            'endian': pwny.Target.Endian.little,
            'abi_version': 0,
            'entry': 4198222,
            'flags': 0,
            'hsize': 64,
            'osabi': pwny.ELF.OSABI.linux,
            'phentsize': 56,
            'phnum': 6,
            'phoff': 64,
            'shentsize': 64,
            'shnum': 31,
            'shoff': 794024,
            'shstrndx': 28,
            'type': pwny.ELF.Type.executable,
        },
    },
    {
        'data': b'\x7fELF\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x03\x00\x01\x00\x00\x00 \x83\x04\x08'
                b'4\x00\x00\x00L\x11\x00\x00\x00\x00\x00\x004\x00 \x00\t\x00(\x00\x1e\x00\x1b\x00\x06\x00\x00\x004\x00'
                b'\x00\x004\x80\x04\x08',
        'header': {
            'arch':pwny.Target.Arch.x86,
            'bits': 32,
            'endian': pwny.Target.Endian.little,
            'abi_version': 0,
            'entry': 134513440,
            'flags': 0,
            'hsize': 52,
            'osabi': pwny.ELF.OSABI.system_v,
            'phentsize': 32,
            'phnum': 9,
            'phoff': 52,
            'shentsize': 40,
            'shnum': 30,
            'shoff': 4428,
            'shstrndx': 27,
            'type': pwny.ELF.Type.executable,
        },
    },
    {
        'data': b'\x7fELF\x01\x01\x01\x03\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x03\x00\x01\x00\x00\x00\n\x8d\x04\x08'
                b'4\x00\x00\x00\xf0 \n\x00\x00\x00\x00\x004\x00 \x00\x06\x00(\x00\x1f\x00\x1c\x00\x01\x00\x00\x00\x00'
                b'\x00\x00\x00\x00\x80\x04\x08',
        'header': {
            'arch': pwny.Target.Arch.x86,
            'bits': 32,
            'endian': pwny.Target.Endian.little,
            'abi_version': 0,
            'entry': 134515978,
            'flags': 0,
            'hsize': 52,
            'osabi': pwny.ELF.OSABI.linux,
            'phentsize': 32,
            'phnum': 6,
            'phoff': 52,
            'sections': [],
            'shentsize': 40,
            'shnum': 31,
            'shoff': 663792,
            'shstrndx': 28,
            'strings': None,
            'type': pwny.ELF.Type.executable,
        },
    }
]


def test_elf_header_parse():
    for header in headers:
        yield check_elf_header_parse, header['data'], header['header']


def check_elf_header_parse(data, values):
    elf = pwny.ELF()
    elf._parse_header(data)
    for key, value in values.items():
        assert getattr(elf, key, value) == value, '%s != %r (%r)' % (key, value, getattr(elf, key))


def test_elf_parse_section_invalid_type():
    section_fmt = 'IIIIIIIIII'
    section_fmt_size = pwny.pack_size(section_fmt)
    section = pwny.pack(
        section_fmt,
        1,
        0x5fffffff,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        section_fmt_size,
    )
    elf = pwny.ELF()
    elf.bits = 32
    section = elf.SectionHeader(elf, section)
    assert section.type == pwny.ELF.SectionHeader.Type.unknown


def test_elf_parse_file_32():
    b = six.BytesIO()

    header_fmt = '4sBBBBB7sHHIIIIIHHHHHH'
    header_fmt_size = pwny.pack_size(header_fmt)

    section_fmt = 'IIIIIIIIII'
    section_fmt_size = pwny.pack_size(section_fmt)

    strings_section = b'\x00strings\x00'

    b.write(pwny.pack(
        header_fmt,
        b'\x7fELF',
        1,
        1,
        1,
        pwny.ELF.OSABI.linux.value,
        0,
        b'\x00' * 7,
        pwny.ELF.Type.executable.value,
        pwny.ELF.Machine.i386.value,
        1,
        0,
        0,
        header_fmt_size,
        0,
        header_fmt_size,
        0,
        0,
        section_fmt_size,
        1,
        0,
    ))

    b.write(pwny.pack(
        section_fmt,
        1,
        pwny.ELF.SectionHeader.Type.null.value,
        0,
        0,
        header_fmt_size + section_fmt_size,
        len(strings_section),
        0,
        0,
        0,
        section_fmt_size,
    ))

    b.write(strings_section)
    b.seek(0)

    elf = pwny.ELF(b)

    for key, value in {
        'arch': pwny.Target.Arch.x86,
        'bits': 32,
        'endian': pwny.Target.Endian.little,
        'abi_version': 0,
        'entry': 0,
        'flags': 0,
        'hsize': header_fmt_size,
        'osabi': pwny.ELF.OSABI.linux,
        'phentsize': 0,
        'phnum': 0,
        'phoff': 0,
        'shentsize': section_fmt_size,
        'shnum': 1,
        'shoff': header_fmt_size,
        'shstrndx': 0,
        'type': pwny.ELF.Type.executable,
    }.items():
        assert getattr(elf, key) == value, '%s != %r (%r)' % (key, value, getattr(elf, key))

    assert len(elf.section_headers) == 1


def test_elf_parse_file_64():
    b = six.BytesIO()

    header_fmt = '4sBBBBB7sHHIQQQIHHHHHH'
    header_fmt_size = pwny.pack_size(header_fmt)

    section_fmt = 'IIQQQQIIQQ'
    section_fmt_size = pwny.pack_size(section_fmt)

    strings_section = b'\x00strings\x00'

    b.write(pwny.pack(
        header_fmt,
        b'\x7fELF',
        2,
        1,
        1,
        pwny.ELF.OSABI.linux.value,
        0,
        b'\x00' * 7,
        pwny.ELF.Type.executable.value,
        pwny.ELF.Machine.x86_64.value,
        1,
        0,
        0,
        header_fmt_size,
        0,
        header_fmt_size,
        0,
        0,
        section_fmt_size,
        1,
        0,
    ))

    b.write(pwny.pack(
        section_fmt,
        1,
        pwny.ELF.SectionHeader.Type.null.value,
        0,
        0,
        header_fmt_size + section_fmt_size,
        len(strings_section),
        0,
        0,
        0,
        section_fmt_size,
    ))

    b.write(strings_section)
    b.seek(0)

    elf = pwny.ELF(b)

    for key, value in {
        'arch': pwny.Target.Arch.x86,
        'bits': 64,
        'endian': pwny.Target.Endian.little,
        'abi_version': 0,
        'entry': 0,
        'flags': 0,
        'hsize': header_fmt_size,
        'osabi': pwny.ELF.OSABI.linux,
        'phentsize': 0,
        'phnum': 0,
        'phoff': 0,
        'shentsize': section_fmt_size,
        'shnum': 1,
        'shoff': header_fmt_size,
        'shstrndx': 0,
        'type': pwny.ELF.Type.executable,
    }.items():
        assert getattr(elf, key) == value, '%s != %r' % (key, value)

    assert len(elf.section_headers) == 1


def test_elf_parse_file_open():
    b = six.BytesIO()

    header_fmt = '4sBBBBB7sHHIIIIIHHHHHH'
    header_fmt_size = pwny.pack_size(header_fmt)

    b.write(pwny.pack(
        header_fmt,
        b'\x7fELF',
        1,
        1,
        1,
        pwny.ELF.OSABI.linux.value,
        0,
        b'\x00' * 7,
        pwny.ELF.Type.executable.value,
        pwny.ELF.Machine.i386.value,
        1,
        0,
        0,
        header_fmt_size,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ))

    b.seek(0)

    with mock.patch('pwnypack.elf.open', create=True) as mock_open:
        mock_open.return_value = b
        pwny.ELF('test.elf')



import argparse
import pwnypack.main


__all__ = []


BANNER = 'Welcome to the interactive pwnypack shell.\n'


try:
    import bpython
    have_bpython = True
except ImportError:
    have_bpython = False


try:
    import IPython
    have_IPython = True
except ImportError:
    have_IPython = False


@pwnypack.main.register()
def shell(_parser, cmd, args):  # pragma: no cover
    """
    Start an interactive python interpreter with pwny imported globally.
    """

    parser = argparse.ArgumentParser(
        prog=_parser.prog,
        description=_parser.description,
    )

    group = parser.add_mutually_exclusive_group()
    group.set_defaults(shell=have_bpython and 'bpython' or (have_IPython and 'ipython' or 'python'))
    if have_bpython:
        group.add_argument(
            '--bpython',
            action='store_const',
            dest='shell',
            const='bpython',
            help='Use the bpython interpreter'
        )
    if have_IPython:
        group.add_argument(
            '--ipython',
            action='store_const',
            dest='shell',
            const='ipython',
            help='Use the IPython interpreter'
        )
    group.add_argument(
        '--python',
        action='store_const',
        dest='shell',
        const='python',
        help='Use the default python interpreter'
    )

    args = parser.parse_args(args)

    import pwny
    pwny_locals = dict(
        (key, getattr(pwny, key))
        for key in dir(pwny)
        if not key.startswith('__') and not key == 'shell'
    )

    if args.shell == 'bpython':
        from bpython import embed
        embed(pwny_locals, banner=BANNER)
    elif args.shell == 'ipython':
        from IPython import start_ipython
        from IPython.config import get_config

        config = get_config()
        config.InteractiveShell.confirm_exit = False
        start_ipython(
            argv=[],
            config=config,
            user_ns=pwny_locals,
        )
    else:
        import code
        code.interact(BANNER, local=pwny_locals)



from __future__ import print_function
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import io
import os
import sys
import argparse
import six
import pwnypack.target


__all__ = []


MAIN_FUNCTIONS = OrderedDict()


def register(name=None, symlink=True):
    def wrapper(f):
        if name is None:
            f_name = f.__name__
        else:
            f_name = name
        if f_name in MAIN_FUNCTIONS:
            raise ValueError('Duplicate application %s' % f_name)
        MAIN_FUNCTIONS[f_name] = {
            'callable': f,
            'symlink': symlink,
        }
        return f
    return wrapper


def binary_value_or_stdin(value):
    """
    Return fsencoded value or read raw data from stdin if value is None.
    """
    if value is None:
        reader = io.open(sys.stdin.fileno(), mode='rb', closefd=False)
        return reader.read()
    elif six.PY3:
        return os.fsencode(value)
    else:
        return value


def string_value_or_stdin(value):
    """
    Return value or read string from stdin if value is None.
    """
    if value is None:
        return sys.stdin.read()
    else:
        return value


def add_target_arguments(parser):
    parser.add_argument(
        '--arch', '-a',
        choices=[v.value for v in pwnypack.target.Target.Arch.__members__.values()],
        default=None,
        help='the target architecture',
    )
    parser.add_argument(
        '--bits', '-b',
        type=int,
        choices=[v.value for v in pwnypack.target.Target.Bits.__members__.values()],
        default=None,
        help='the target word size',
    )
    parser.add_argument(
        '--endian', '-e',
        choices=pwnypack.target.Target.Endian.__members__.keys(),
        default=None,
        help='the target endianness',
    )


def target_from_arguments(args):
    if args.endian is not None:
        endian = pwnypack.target.Target.Endian.__members__[args.endian]
    else:
        endian = None
    return pwnypack.target.Target(arch=args.arch, bits=args.bits, endian=endian)


@register(symlink=False)
def symlink(parser, cmd, args):
    """
    Set up symlinks for (a subset of) the pwny apps.
    """

    parser.add_argument(
        'apps',
        nargs=argparse.REMAINDER,
        help='Which apps to create symlinks for.'
    )
    args = parser.parse_args(args)

    base_dir, pwny_main = os.path.split(sys.argv[0])

    for app_name, config in MAIN_FUNCTIONS.items():
        if not config['symlink'] or (args.apps and app_name not in args.apps):
            continue
        dest = os.path.join(base_dir, app_name)
        if not os.path.exists(dest):
            print('Creating symlink %s' % dest)
            os.symlink(pwny_main, dest)
        else:
            print('Not creating symlink %s (file already exists)' % dest)


def main():
    def usage():
        global MAIN_FUNCTIONS
        print('Welcome to pwny!')
        print()
        print('Available apps:')
        longest_app_name = max(len(app) for app in MAIN_FUNCTIONS)
        for app, config in MAIN_FUNCTIONS.items():
            f = config['callable']
            fmt = ' - %%-%ds   %%s' % longest_app_name
            print(fmt % (app, f.__doc__.strip().split('\n')[0]))
        print()
        sys.exit(1)

    # Import everything so all main functions are registered.
    import pwny

    global MAIN_FUNCTIONS

    app = os.path.basename(sys.argv[0])
    app_args = sys.argv[1:]

    if app not in MAIN_FUNCTIONS:
        if len(sys.argv) < 2 or app_args[0] not in MAIN_FUNCTIONS:
            usage()
        prog = '%s %s' % (app, app_args[0])
        app, app_args = app_args[0], app_args[1:]
    else:
        prog = app

    f = MAIN_FUNCTIONS[app]['callable']
    parser = argparse.ArgumentParser(
        prog=prog,
        description=f.__doc__,
    )
    parser.add_argument(
        '-f', '--format',
        dest='format',
        choices=['raw', 'hex', 'py', 'sh', 'b64'],
        default='raw',
        help='set output format'
    )
    parser.add_argument(
        '-n', '--no-newline',
        dest='no_newline',
        action='store_const',
        const=True,
        help='inhibit newline after output'
    )

    output = f(parser, app, app_args)

    if output is not None:
        if six.PY3:
            output_bytes = os.fsencode(output)
        else:
            output_bytes = output

        args = parser.parse_args(app_args)
        if args.no_newline:
            end = ''
        else:
            end = '\n'

        if args.format == 'raw':
            if isinstance(output, six.binary_type):
                writer = io.open(sys.stdout.fileno(), mode='wb', closefd=False)
                writer.write(output_bytes)
                writer.write(end)
                writer.close()
            else:
                print(output, end=end)
        elif args.format == 'hex':
            print(pwny.enhex(output_bytes), end=end)
        elif args.format == 'py':
            print(repr(output), end=end)
        elif args.format == 'sh':
            r = repr(output)
            if r.startswith('b\''):
                r = r[1:]
            print('$' + r, end=end)
        elif args.format == 'b64':
            print(pwny.enb64(output_bytes), end=end)



"""
The util module contains various utility functions.
"""

from __future__ import print_function
import argparse
from six.moves import range, cPickle
import re
import binascii
import sys
import pwnypack.main
import pwnypack.codec


__all__ = [
    'cycle',
    'cycle_find',
    'reghex',
    'pickle_call',
]


def deBruijn(n, k):
    """
    An implementation of the FKM algorithm for generating the de Bruijn
    sequence containing all k-ary strings of length n, as described in
    "Combinatorial Generation" by Frank Ruskey.
    """

    a = [ 0 ] * (n + 1)

    def gen(t, p):
        if t > n:
            for v in a[1:p + 1]:
                yield v
        else:
            a[t] = a[t - p]
         
            for v in gen(t + 1, p):
                yield v
         
            for j in range(a[t - p] + 1, k):
                a[t] = j
                for v in gen(t + 1, t):
                    yield v

    return gen(1, 1)


def cycle(length, width=4):
    """
    Generate a de Bruijn sequence of a given length (and width). A de Bruijn
    sequence is a set of varying repetitions where each sequence of *n*
    characters is unique within the sequence. This type of sequence can be
    used to easily find the offset to the return pointer when exploiting a
    buffer overflow.

    Args:
        length(int): The length of the sequence to generate.
        width(int): The width of each element in the sequence.

    Returns:
        str: The sequence.

    Example:
        >>> from pwny import *
        >>> cycle(80)
        AAAABAAACAAADAAAEAAAFAAAGAAAHAAAIAAAJAAAKAAALAAAMAAANAAAOAAAPAAAQAAARAAASAAATAAA
    """

    iter = deBruijn(width, 26)
    return ''.join([chr(ord('A') + next(iter)) for i in range(length)])


def cycle_find(key, width=4):
    """
    Given an element of a de Bruijn sequence, find its index in that sequence.

    Args:
        key(str): The piece of the de Bruijn sequence to find.
        width(int): The width of each element in the sequence.

    Returns:
        int: The index of ``key`` in the de Bruijn sequence.
    """

    key_len = len(key)
    buf = ''

    it = deBruijn(width, 26)

    for i in range(key_len):
        buf += chr(ord('A') + next(it))

    if buf == key:
        return 0

    for i, c in enumerate(it):
        buf = buf[1:] + chr(ord('A') + c)
        if buf == key:
            return i + 1

    return -1


REGHEX_PATTERN = r'(([a-fA-F0-9]{2})|(([?.])(\{\d+\})?)|(\*|\+)|\s+)'
reghex_check = re.compile(REGHEX_PATTERN + '+')
reghex_regex = re.compile(REGHEX_PATTERN)


def reghex(pattern):
    """
    Compile a regular hexpression (a short form regular expression subset
    specifically designed for searching for binary strings).

    A regular hexpression consists of hex tuples interspaced with control
    characters. The available control characters are:

    - ``?``: Any byte (optional).
    - ``.``: Any byte (required).
    - ``?{n}``: A set of 0 up to *n* bytes.
    - ``.{n}``: A set of exactly *n* bytes.
    - ``*``: Any number of bytes (or no bytes at all).
    - ``+``: Any number of bytes (at least one byte).

    Args:
        pattern(str): The reghex pattern.

    Returns:
        regexp: A regular expression as returned by ``re.compile()``.
    """

    if not reghex_check.match(pattern):
        raise SyntaxError('Invalid reghex pattern.')

    b_pattern = b''

    for match in reghex_regex.finditer(pattern):
        _, match_hex, _, match_char, match_char_len, match_star_plus = match.groups()
        if match_hex:
            b_pattern += pwnypack.codec.dehex(match_hex)
        elif match_char:
            if match_char == '?':
                if match_char_len is None:
                    b_pattern += b'.?'
                else:
                    b_pattern += ('.{0,%d}?' % int(match_char_len[1:-1])).encode('ascii')
            else:
                if match_char_len is None:
                    b_pattern += b'.'
                else:
                    b_pattern += b'.' * int(match_char_len[1:-1])
        elif match_star_plus:
            b_pattern += b'.' + match_star_plus.encode('ascii') + b'?'

    try:
        return re.compile(b_pattern)
    except (TypeError, binascii.Error, re.error):
        raise SyntaxError('Invalid reghex pattern.')


def pickle_call(func, *args):
    """
    Create a byte sequence which when unpickled calls a callable with given
    arguments.

    Arguments:
        func(callable): The function to call or class to instantiate.
        args(tuple): The arguments to call the callable with.

    Returns:
        bytes: The data that when unpickled calls ``func(*args)``.

    Example:
        >>> from pwny import *
        >>> import pickle
        >>> def hello(arg):
        ...     print('Hello, %s!' % arg)
        ...
        >>> pickle.loads(pickle_call(hello, 'world'))
        Hello, world!
    """

    class CallFunc(object):
        def __reduce__(self):
            return func, tuple(args)
    return cPickle.dumps(CallFunc(), 0)


@pwnypack.main.register('cycle')
def cycle_app(parser, cmd, args):  # pragma: no cover
    """
    Generate a de Bruijn sequence of a given length.
    """

    parser.add_argument('-w', '--width', type=int, default=4, help='the length of the cycled value')
    parser.add_argument('length', type=int, help='the cycle length to generate')
    args = parser.parse_args(args)
    return cycle(args.length, args.width)


@pwnypack.main.register('cycle-find')
def cycle_find_app(_parser, cmd, args):  # pragma: no cover
    """
    Find the first position of a value in a de Bruijn sequence.
    """

    parser = argparse.ArgumentParser(
        prog=_parser.prog,
        description=_parser.description,
    )
    parser.add_argument('-w', '--width', type=int, default=4, help='the length of the cycled value')
    parser.add_argument('value', help='the value to determine the position of, read from stdin if missing', nargs='?')
    args = parser.parse_args(args)
    index = cycle_find(pwnypack.main.string_value_or_stdin(args.value), args.width)
    if index == -1:
        print('Not found.')
        sys.exit(1)
    else:
        print('Found at position: %d' % index)



import pwny


def test_xor_int():
    assert pwny.xor(61, b'fooo') == b'[RRR'


def test_xor_str():
    assert pwny.xor(b'abcd', b'fooo') == b'\x07\r\x0c\x0b'
    assert pwny.xor(b'abcd', b'fooofooo') == b'\x07\r\x0c\x0b\x07\r\x0c\x0b'


def test_rot13():
    assert pwny.rot13('whax') == 'junk'


def test_caesar():
    assert pwny.caesar(1, 'abcXYZ') == 'bcdYZA'


def test_enhex():
    assert pwny.enhex(b'ABCD') == '41424344'


def test_dehex():
    assert pwny.dehex('41424344') == b'ABCD'


def test_enb64():
    assert pwny.enb64(b'ABCD') == 'QUJDRA=='


def test_deb64():
    assert pwny.deb64('QUJDRA==') == b'ABCD'


def test_deurlform():
    assert pwny.deurlform('foo=bar&baz=quux&baz=corge') == {'foo': ['bar'], 'baz': ['quux', 'corge']}


def test_enurlform():
    assert pwny.enurlform((('foo', 'bar'), ('baz', ['quux', 'corge']))) == 'foo=bar&baz=quux&baz=corge'


def test_enurlquote():
    assert pwny.enurlquote('Foo Bar/Baz') == 'Foo%20Bar/Baz'


def test_enurlquote_plus():
    assert pwny.enurlquote('Foo Bar/Baz', plus=True) == 'Foo+Bar%2FBaz'


def test_deurlquote():
    assert pwny.deurlquote('Foo%20Bar%2FBaz') == 'Foo Bar/Baz'


def test_deurlquote_no_plus():
    assert pwny.deurlquote('Foo+Bar%2FBaz') == 'Foo+Bar/Baz'


def test_deurlquote_plus():
    assert pwny.deurlquote('Foo+Bar%2FBaz', True) == 'Foo Bar/Baz'


def test_frequency():
    assert pwny.frequency('ABCD') == {'A': 1, 'B': 1, 'C': 1, 'D': 1}



"""
The ROP module contains a function to find gadgets in ELF binaries that can
be used to create ROP chains.
"""

from __future__ import print_function
import argparse
import re
import capstone
import six
import sys
import pwnypack.codec
import pwnypack.elf
import pwnypack.main
import pwnypack.asm
import pwnypack.target
import pwnypack.util


__all__ = [
    'find_gadget',
]


INVALID_GROUPS = set((capstone.CS_GRP_CALL, capstone.CS_GRP_JUMP))


def find_gadget(elf, gadget, align=1, unique=True):
    """
    Find a ROP gadget in a the executable sections of an ELF executable or
    library. The ROP gadget can be either a set of bytes for an exact match
    or a (bytes) regular expression. Once it finds gadgets, it uses the
    capstone engine to verify if the gadget consists of valid instructions
    and doesn't contain any call or jump instructions.

    Args:
        elf(:class:`~pwnypack.elf.ELF`): The ELF instance to find a gadget in.
        gadget(bytes or regexp): The gadget to find.
        align(int): Make sure the gadget starts at a multiple of this number
        unique(bool): If true, only unique gadgets are returned.

    Returns:
        dict: A dictionary containing a description of the found
            gadget. Contains the following fields:

            - section: The section the gadget was found in.
            - offset: The offset inside the segment the gadget was found at.
            - addr: The virtual memory address the gadget will be located at.
            - gadget: The machine code of the found gadget.
            - asm: A list of disassembled instructions.

    """

    if not isinstance(elf, pwnypack.elf.ELF):
        elf = pwnypack.elf.ELF(elf)

    matches = []
    gadgets = []

    if isinstance(gadget, six.binary_type):
        gadget = re.compile(re.escape(gadget))

    for section in elf.section_headers:
        if section.type != section.Type.progbits:
            continue

        for match in gadget.finditer(section.content):
            match_index = match.start()
            if match_index % align != 0:
                continue

            match_gadget = match.group()

            if match_gadget in gadgets:
                continue

            match_addr = section.addr + match_index

            md = pwnypack.asm.prepare_capstone(syntax=pwnypack.asm.AsmSyntax.intel, target=elf)
            md.detail = True
            match_asm = []

            for insn in md.disasm(match_gadget, match_addr):
                if insn.id == capstone.CS_OP_INVALID or set(insn.groups) & INVALID_GROUPS:
                    # Don't try to disassemble this particular gadget again.
                    gadgets.append(match_gadget)
                    break
                match_asm.append((insn.mnemonic + ' ' + insn.op_str).strip())
            else:
                matches.append({
                    'section': section,
                    'offset': match_index,
                    'addr': match_addr,
                    'gadget': match_gadget,
                    'asm': match_asm,
                })
                if unique:
                    gadgets.append(match_gadget)

    return matches


@pwnypack.main.register('gadget')
def gadget_app(_parser, cmd, args):  # pragma: no cover
    """
    Find ROP gadgets in an ELF binary.
    """

    parser = argparse.ArgumentParser(
        prog=_parser.prog,
        description=_parser.description,
    )
    parser.add_argument('file', help='ELF file to find gadgets in')
    parser.add_argument('gadget', help='the assembler source or reghex expression')
    parser.add_argument(
        '--reghex', '-r',
        dest='mode',
        action='store_const',
        const='reghex',
        help='use reghex expression (hex bytes interspaced with ? for wildcard)',
    )
    parser.add_argument(
        '--asm', '-a',
        dest='mode',
        action='store_const',
        const='asm',
        help='use assembler expression (separate lines with semi-colon)',
    )
    parser.add_argument(
        '--all', '-l',
        dest='unique',
        action='store_const',
        const=False,
        default=True,
        help='also show non-unique gadgets',
    )
    args = parser.parse_args(args)

    if args.mode is None:
        try:
            pwnypack.util.reghex(args.gadget)
            args.mode = 'reghex'
        except SyntaxError:
            args.mode = 'asm'

    elf = pwnypack.elf.ELF(args.file)

    if args.mode == 'reghex':
        try:
            gadget = pwnypack.util.reghex(args.gadget)
        except SyntaxError:
            print('Invalid reghex pattern.')
            sys.exit(1)
    else:
        try:
            gadget = pwnypack.asm.asm(args.gadget.replace(';', '\n'), target=elf)
        except SyntaxError as e:
            print('Could not assemble:', e.msg)
            sys.exit(1)

    matches = find_gadget(
        elf,
        gadget,
        unique=args.unique
    )

    if not matches:
        print('No gadgets found.', file=sys.stdout)
        return

    longest_gadget = max(len(m['gadget']) for m in matches)
    fmt = '  0x%%0%dx: [ %%-%ds ] %%s' % (elf.bits / 4, longest_gadget * 3 - 1)

    current_section = None

    for match in matches:
        if match['section'].name != current_section:
            if current_section is not None:
                print()
            print('Section: %s' % match['section'].name)
            current_section = match['section'].name

        hex_gadget = pwnypack.codec.enhex(match['gadget'])
        print(fmt % (
            match['addr'],
            ' '.join(
                hex_gadget[i:i+2]
                for i in range(0, len(hex_gadget), 2)
            ),
            ' ; '.join(match['asm'])
        ))

    print()



import pwnypack.target
import struct


__all__ = [
    'pack',
    'unpack',
    'pack_size',
    'P',
    'p',
    'U',
    'u',
]


def pack(fmt, *args, **kwargs):
    """pack(fmt, v1, v2, ..., endian=None, target=None)

    Return a string containing the values v1, v2, ... packed according to the
    given format. The actual packing is performed by ``struct.pack`` but the
    byte order will be set according to the given `endian`, `target` or
    byte order of the global target.

    Args:
        fmt(str): The format string.
        v1,v2,...: The values to pack.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.

    Returns:
        bytes: The provided values packed according to the format.
    """

    endian, target = kwargs.get('endian'), kwargs.get('target')
    endian = endian if endian is not None else target.endian if target is not None else pwnypack.target.target.endian
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Target.Endian.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Target.Endian.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.pack(fmt, *args)


def unpack(fmt, data, endian=None, target=None):
    """
    Unpack the string (presumably packed by pack(fmt, ...)) according to the
    given format. The actual unpacking is performed by ``struct.unpack``
    but the byte order will be set according to the given `endian`, `target`
    or byte order of the global target.

    Args:
        fmt(str): The format string.
        data(bytes): The data to unpack.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.

    Returns:
        list: The unpacked values according to the format.
    """

    endian = endian if endian is not None else target.endian if target is not None else pwnypack.target.target.endian
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Target.Endian.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Target.Endian.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.unpack(fmt, data)


def pack_size(fmt, endian=None, target=None):
    endian = endian if endian is not None else target.endian if target is not None else pwnypack.target.target.endian
    if fmt and fmt[0] not in '@=<>!':
        if endian is pwnypack.target.Target.Endian.little:
            fmt = '<' + fmt
        elif endian is pwnypack.target.Target.Endian.big:
            fmt = '>' + fmt
        else:
            raise NotImplementedError('Unsupported endianness: %s' % endian)
    return struct.calcsize(fmt)


def _pack_closure(fmt):
    return lambda value, endian=None, target=None: pack(fmt, value, endian=endian, target=target)


def _unpack_closure(fmt):
    return lambda data, endian=None, target=None: unpack(fmt, data, endian=endian, target=target)[0]


for _w, _f in ((8, 'b'), (16, 'h'), (32, 'l'), (64, 'q')):
    locals().update({
        'p%d' % _w: _pack_closure(_f),
        'P%d' % _w: _pack_closure(_f.upper()),
        'u%d' % _w: _unpack_closure(_f),
        'U%d' % _w: _unpack_closure(_f.upper()),
    })

    locals()['p%d' % _w].__doc__ = '''Pack signed %d bit integer. Alias for ``pack('%s', ...)``.''' % (_w, _f)
    locals()['P%d' % _w].__doc__ = '''Pack unsigned %d bit integer. Alias for ``pack('%s', ...)``.''' % (_w, _f.upper())
    locals()['u%d' % _w].__doc__ = '''Unpack signed %d bit integer. Alias for ``unpack('%s', ...)``.''' % (_w, _f)
    locals()['U%d' % _w].__doc__ = '''Unpack unsigned %d bit integer. Alias for ``unpack('%s', ...)``.''' % \
                                   (_w, _f.upper())

    __all__.extend([
        'p%d' % _w,
        'P%d' % _w,
        'u%d' % _w,
        'U%d' % _w,
    ])
del _w, _f, _pack_closure, _unpack_closure


def P(value, bits=None, endian=None, target=None):
    """
    Pack an unsigned pointer for a given target.

    Args:
        value(int): The value to pack.
        bits(:class:`~pwnypack.target.Target.Bits`): Override the default
            word size. If ``None`` it will look at the word size of
            ``target``.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.
    """

    bits = pwnypack.target.Target.Bits(bits) if bits is not None else target.bits if target is not None else \
        pwnypack.target.target.bits
    return globals()['P%d' % bits.value](value, endian=endian, target=target)


def p(value, bits=None, endian=None, target=None):
    """
    Pack a signed pointer for a given target.

    Args:
        value(int): The value to pack.
        bits(:class:`pwnypack.target.Target.Bits`): Override the default
            word size. If ``None`` it will look at the word size of
            ``target``.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.
    """

    bits = pwnypack.target.Target.Bits(bits) if bits is not None else target.bits if target is not None else \
        pwnypack.target.target.bits
    return globals()['p%d' % bits.value](value, endian=endian, target=target)


def U(data, bits=None, endian=None, target=None):
    """
    Unpack an unsigned pointer for a given target.

    Args:
        data(bytes): The data to unpack.
        bits(:class:`pwnypack.target.Target.Bits`): Override the default
            word size. If ``None`` it will look at the word size of
            ``target``.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.

    Returns:
        int: The pointer value.
    """

    bits = pwnypack.target.Target.Bits(bits) if bits is not None else target.bits if target is not None else \
        pwnypack.target.target.bits
    return globals()['U%d' % bits.value](data, endian=endian, target=target)


def u(data, bits=None, endian=None, target=None):
    """
    Unpack a signed pointer for a given target.

    Args:
        data(bytes): The data to unpack.
        bits(:class:`pwnypack.target.Target.Bits`): Override the default
            word size. If ``None`` it will look at the word size of
            ``target``.
        endian(:class:`~pwnypack.target.Target.Endian`): Override the default
            byte order. If ``None``, it will look at the byte order of
            the ``target`` argument.
        target(:class:`~pwnypack.target.Target`): Override the default byte
            order. If ``None``, it will look at the byte order of
            the global :data:`~pwnypack.target.target`.

    Returns:
        int: The pointer value.
    """

    bits = pwnypack.target.Target.Bits(bits) if bits is not None else target.bits if target is not None else \
        pwnypack.target.target.bits
    return globals()['u%d' % bits.value](data, endian=endian, target=target)



import mock
from nose.tools import raises
import pwny


def test_default_arch_x86():
    with mock.patch('platform.machine') as platform_mock:
        platform_mock.return_value = 'i386'
        assert pwny.Target().arch is pwny.Target.Arch.x86


def test_default_arch_x86_64():
    with mock.patch('platform.machine') as platform_mock:
        platform_mock.return_value = 'x86_64'
        assert pwny.Target().arch is pwny.Target.Arch.x86


def test_default_arch_unknown():
    with mock.patch('platform.machine') as platform_mock:
        platform_mock.return_value = 'unknown'
        assert pwny.Target().arch is pwny.Target.Arch.unknown


def test_default_arch_32bit():
    with mock.patch('platform.architecture') as platform_mock:
        platform_mock.return_value = ('32bit',)
        assert pwny.Target().bits is pwny.Target.Bits.bits_32


def test_default_arch_64bit():
    with mock.patch('platform.architecture') as platform_mock:
        platform_mock.return_value = ('64bit',)
        assert pwny.Target().bits is pwny.Target.Bits.bits_64


def test_set_arch():
    with mock.patch('platform.architecture') as platform_mock:
        platform_mock.return_value = ('64bit',)
        target = pwny.Target(arch=pwny.Target.Arch.x86)
        assert target.arch is pwny.Target.Arch.x86


def test_default_endian():
    assert pwny.Target().endian is pwny.Target.Endian.little


def test_set_endian():
    target = pwny.Target(arch=pwny.Target.Arch.unknown, endian=pwny.Target.Endian.big)
    assert target.endian is pwny.Target.Endian.big


def test_default_bits_x86():
    target = pwny.Target(arch=pwny.Target.Arch.x86)
    assert target.bits == 32


@raises(NotImplementedError)
def test_default_bits_unsupported():
    target = pwny.Target(arch=pwny.Target.Arch.unknown)
    _ = target.bits


def test_set__bits():
    target = pwny.Target(arch=pwny.Target.Arch.x86, bits=64)
    assert target.bits == 64


@raises(ValueError)
def test_set_invalid_bits():
    pwny.Target(bits=33)


def test_target_assume():
    target = pwny.Target()
    target.assume(pwny.Target(arch=pwny.Target.Arch.arm, endian=pwny.Target.Endian.little, bits=64, mode=2))
    assert target.arch is pwny.Target.Arch.arm and \
           target.endian == pwny.Target.Endian.little and \
           target.bits == 64 and \
           target.mode == 2



"""
This module contains functions to assemble and disassemble code for a given
target platform.

Currently, the only supported architecture is
:attr:`~pwnypack.target.Target.Arch.x86` (both 32 and 64 bits variants).
Assembly is performed by the *nasm* assembler (only supports
:attr:`~AsmSyntax.nasm` syntax). Disassembly is performed by *ndisasm*
(:attr:`~AsmSyntax.nasm` syntax) or *capstone*
(:attr:`~AsmSyntax.intel` & :attr:`~AsmSyntax.att` syntax).
"""

from __future__ import print_function
import argparse
import os
import subprocess
import sys
import capstone
from enum import IntEnum
from pwnypack.elf import ELF
import pwnypack.target
import pwnypack.main
import pwnypack.codec
import tempfile


__all__ = [
    'AsmSyntax',
    'asm',
    'disasm',
]


class AsmSyntax(IntEnum):
    """
    This enumeration is used to specify the assembler syntax.
    """

    nasm = 0   #: Netwide assembler syntax
    intel = 1  #: Intel assembler syntax
    att = 2    #: AT&T assembler syntax


def asm(code, addr=0, syntax=AsmSyntax.nasm, target=None):
    """asm(code, addr=0, syntax=AsmSyntax.nasm, target=None)

    Assemble statements into machine readable code.

    Args:
        code(str): The statements to assemble.
        addr(int): The memory address where the code will run.
        syntax(AsmSyntax): The input assembler syntax.
        target(~pwnypack.target.Target): The target architecture. The
            global target is used if this argument is ``None``.

    Returns:
        bytes: The assembled machine code.

    Raises:
        SyntaxError: If the assembler statements are invalid.
        NotImplementedError: In an unsupported target platform is specified.

    Example:
        >>> from pwny import *
        >>> asm('''
        ...     pop rdi
        ...     ret
        ... ''', target=Target(arch=Target.Arch.x86, bits=64))
        b'_\\xc3'
    """

    if target is None:
        target = pwnypack.target.target

    if syntax is AsmSyntax.nasm:
        if target.arch is not pwnypack.target.Target.Arch.x86:
            raise NotImplementedError('nasm only supports x86 target platforms.')

        with tempfile.NamedTemporaryFile() as tmp_asm:
            tmp_asm.write(('bits %d\norg %d\n%s' % (target.bits.value, addr, code)).encode('utf-8'))
            tmp_asm.flush()

            tmp_bin_fd, tmp_bin_name = tempfile.mkstemp()
            os.close(tmp_bin_fd)

            try:
                p = subprocess.Popen(
                    [
                        'nasm',
                        '-o', tmp_bin_name,
                        '-f', 'bin',
                        tmp_asm.name,
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                stdout, stderr = p.communicate()

                if p.returncode:
                    raise SyntaxError(stderr.decode('utf-8'))

                tmp_bin = open(tmp_bin_name, 'rb')
                result = tmp_bin.read()
                tmp_bin.close()
                return result
            finally:
                try:
                    os.unlink(tmp_bin_name)
                except OSError:
                    pass
    else:
        raise NotImplementedError('Unsupported syntax for host platform.')


def prepare_capstone(syntax=AsmSyntax.att, target=None):
    """
    Prepare a capstone disassembler instance for a given target and syntax.

    Args:
        syntax(AsmSyntax): The assembler syntax (Intel or AT&T).
        target(~pwnypack.target.Target): The target to create a disassembler
            instance for. The global target is used if this argument is
            ``None``.

    Returns:
        An instance of the capstone disassembler.

    Raises:
        NotImplementedError: If the specified target isn't supported.
    """

    if target is None:
        target = pwnypack.target.target

    if target.arch == pwnypack.target.Target.Arch.x86:
        if target.bits is pwnypack.target.Target.Bits.bits_32:
            md = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_32)
        else:
            md = capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64)
    else:
        raise NotImplementedError('Only x86 is currently supported.')

    md.skipdata = True

    if syntax is AsmSyntax.att:
        md.syntax = capstone.CS_OPT_SYNTAX_ATT
    elif syntax is AsmSyntax.intel:
        md.skipdata_setup(('db', None, None))
    else:
        raise NotImplementedError('capstone engine only implements AT&T and Intel syntax.')

    return md


def disasm(code, addr=0, syntax=AsmSyntax.nasm, target=None):
    """disasm(code, addr=0, syntax=AsmSyntax.nasm, target=None)

    Disassemble machine readable code into human readable statements.

    Args:
        code(bytes): The machine code that is to be disassembled.
        addr(int): The memory address of the code (used for relative
            references).
        syntax(AsmSyntax): The output assembler syntax.
        target(~pwnypack.target.Target): The architecture for which the code
            was written.  The global target is used if this argument is
            ``None``.

    Returns:
        list of str: The disassembled machine code.

    Raises:
        NotImplementedError: In an unsupported target platform is specified.
        RuntimeError: If ndisasm encounters an error.

    Example:
        >>> from pwny import *
        >>> disasm(b'_\\xc3', target=Target(arch=Target.Arch.x86, bits=64))
        ['pop rdi', 'ret']
    """

    if target is None:
        target = pwnypack.target.target

    if syntax is AsmSyntax.nasm:
        if target.arch is not pwnypack.target.Target.Arch.x86:
            raise NotImplementedError('nasm only supports x86.')

        p = subprocess.Popen(
            [
                'ndisasm',
                '-b',
                str(target.bits.value),
                '-o',
                str(addr),
                '-',
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate(code)
        if p.returncode:
            raise RuntimeError(stderr.decode('utf-8'))

        return [
            line.split(None, 2)[2]
            for line in stdout.decode('utf-8').split('\n')
            if line and not line.startswith(' ')
        ]
    elif syntax in (AsmSyntax.intel, AsmSyntax.att):
        md = prepare_capstone(syntax, target)
        statements = []
        total_size = 0
        for (_, size, mnemonic, op_str) in md.disasm_lite(code, addr):
            statements.append((mnemonic + ' ' + op_str).strip())
            total_size += size
        return statements
    else:
        raise NotImplementedError('Unsupported syntax for host platform.')


@pwnypack.main.register('asm')
def asm_app(parser, cmd, args):  # pragma: no cover
    """
    Assemble code from commandline or stdin.

    Please not that all semi-colons are replaced with carriage returns
    unless source is read from stdin.
    """

    parser.add_argument('source', help='the code to assemble, read from stdin if omitted', nargs='?')
    pwnypack.main.add_target_arguments(parser)
    parser.add_argument(
        '--syntax', '-s',
        choices=AsmSyntax.__members__.keys(),
        default='nasm',
    )
    parser.add_argument(
        '--address', '-o',
        type=lambda v: int(v, 0),
        default=0,
        help='the address where the code is expected to run',
    )

    args = parser.parse_args(args)
    target = pwnypack.main.target_from_arguments(args)
    syntax = AsmSyntax.__members__[args.syntax]
    if args.source is None:
        args.source = sys.stdin.read()
    else:
        args.source = args.source.replace(';', '\n')

    return asm(
        args.source,
        syntax=syntax,
        target=target,
    )


@pwnypack.main.register('disasm')
def disasm_app(_parser, cmd, args):  # pragma: no cover
    """
    Disassemble code from commandline or stdin.
    """

    parser = argparse.ArgumentParser(
        prog=_parser.prog,
        description=_parser.description,
    )
    parser.add_argument('code', help='the code to disassemble, read from stdin if omitted', nargs='?')
    pwnypack.main.add_target_arguments(parser)
    parser.add_argument(
        '--syntax', '-s',
        choices=AsmSyntax.__members__.keys(),
        default='nasm',
    )
    parser.add_argument(
        '--address', '-o',
        type=lambda v: int(v, 0),
        default=0,
        help='the address of the disassembled code',
    )
    parser.add_argument(
        '--format', '-f',
        choices=['hex', 'bin'],
        help='the input format (defaults to hex for commandline, bin for stdin)',
    )

    args = parser.parse_args(args)
    target = pwnypack.main.target_from_arguments(args)
    syntax = AsmSyntax.__members__[args.syntax]

    if args.format is None:
        if args.code is None:
            args.format = 'bin'
        else:
            args.format = 'hex'

    if args.format == 'hex':
        code = pwnypack.codec.dehex(pwnypack.main.string_value_or_stdin(args.code))
    else:
        code = pwnypack.main.binary_value_or_stdin(args.code)

    print('\n'.join(disasm(code, args.address, syntax=syntax, target=target)))


@pwnypack.main.register(name='symbol-disasm')
def disasm_symbol_app(_parser, _, args):  # pragma: no cover
    """
    Disassemble a symbol from an ELF file.
    """

    parser = argparse.ArgumentParser(
        prog=_parser.prog,
        description=_parser.description,
    )
    parser.add_argument(
        '--syntax', '-s',
        choices=AsmSyntax.__members__.keys(),
        default='nasm',
    )
    parser.add_argument('file', help='ELF file to extract a symbol from')
    parser.add_argument('symbol', help='the symbol to disassemble')

    args = parser.parse_args(args)
    syntax = AsmSyntax.__members__[args.syntax]
    elf = ELF(args.file)
    symbol = elf.get_symbol(args.symbol)
    print('\n'.join(disasm(symbol.content, symbol.value, syntax=syntax, target=elf)))



"""
The Flow module lets you connect to processes or network services using a
unified API. It is primarily designed for synchronous communication flows.

It is based around the central :class:`Flow` class which uses a ``Channel``
to connect to a process. The :class:`Flow` class then uses the primitives
exposed by the ``Channel`` to provide a high level API for reading/receiving
and writing/sending data.


Examples:
    >>> from pwny import *
    >>> f = Flow.connect_tcp('ced.pwned.systems', 80)
    >>> f.writelines([
    ...     b'GET / HTTP/1.0',
    ...     b'Host: ced.pwned.systems',
    ...     b'',
    ... ])
    >>> line = f.readline().strip()
    >>> print(line == b'HTTP/1.0 200 OK')
    True
    >>> f.until(b'\\r\\n\\r\\n')
    >>> f.read_eof(echo=True)
    ... lots of html ...

    >>> from pwny import *
    >>> f = Flow.execute('cat')
    >>> f.writeline(b'hello')
    >>> f.readline(echo=True)
"""

import subprocess
import sys
import socket
import select
import paramiko


__all__ = [
    'ProcessChannel',
    'SocketChannel',
    'TCPClientSocketChannel',
    'Flow',
]


class ProcessChannel(object):
    """ProcessChannel(executable, argument..., redirect_stderr=False)

    This channel type allows controlling processes. It uses python's
    ``subprocess.Popen`` class to execute a process and allows you to
    communicate with it.

    Args:
        executable(str): The executable to start.
        argument...(list of str): The arguments to pass to the executable.
        redirect_stderr(bool): Whether to also capture the output of stderr.
    """

    def __init__(self, executable, *arguments, **kwargs):
        if kwargs.get('redirect_stderr'):
            stderr = subprocess.STDOUT
        else:
            stderr = None

        self._process = subprocess.Popen(
            (executable,) + tuple(arguments),
            bufsize=0,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=stderr,
        )

    def fileno(self):
        """
        Return the file descriptor number for the stdout channel of this
        process.
        """

        return self._process.stdout.fileno()

    def read(self, n):
        """
        Read *n* bytes from the subprocess' output channel.

        Args:
            n(int): The number of bytes to read.

        Returns:
            bytes: *n* bytes of output.

        Raises:
            EOFError: If the process exited.
        """

        d = b''
        while n:
            try:
                block = self._process.stdout.read(n)
            except ValueError:
                block = None
            if not block:
                self._process.poll()
                raise EOFError('Process ended')
            d += block
            n -= len(block)
        return d

    def write(self, data):
        """
        Write *n* bytes to the subprocess' input channel.

        Args:
            data(bytes): The data to write.

        Raises:
            EOFError: If the process exited.
        """

        self._process.poll()
        if self._process.returncode is not None:
            raise EOFError('Process ended')
        self._process.stdin.write(data)

    def close(self):
        """
        Wait for the subprocess to exit.
        """

        self._process.communicate()

    def kill(self):
        """
        Terminate the subprocess.
        """

        self._process.kill()


class SocketChannel(object):
    """
    This channel type allows controlling sockets.

    Args:
        socket(socket.socket): The (already connected) socket to control.
    """

    def __init__(self, sock):
        self._socket = sock

    def fileno(self):
        """
        Return the file descriptor number for the socket.
        """

        return self._socket.fileno()

    def read(self, n):
        """
        Receive *n* bytes from the socket.

        Args:
            n(int): The number of bytes to read.

        Returns:
            bytes: *n* bytes read from the socket.

        Raises:
            EOFError: If the socket was closed.
        """

        d = b''
        while n:
            try:
                block = self._socket.recv(n)
            except socket.error:
                block = None
            if not block:
                raise EOFError('Socket closed')
            d += block
            n -= len(block)
        return d

    def write(self, data):
        """
        Send *n* bytes to socket.

        Args:
            data(bytes): The data to send.

        Raises:
            EOFError: If the socket was closed.
        """

        while data:
            try:
                n = self._socket.send(data)
            except socket.error:
                n = None
            if not n:
                raise EOFError('Socket closed')
            data = data[n:]

    def close(self):
        """
        Close the socket gracefully.
        """

        self._socket.close()

    def kill(self):
        """
        Shut down the socket immediately.
        """

        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()


class TCPClientSocketChannel(SocketChannel):
    """
    Convenience subclass of :class:`SocketChannel` that allows you to connect
    to a TCP hostname / port pair easily.

    Args:
        host(str): The hostname or IP address to connect to.
        port(int): The port number to connect to.
    """

    def __init__(self, host, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        super(TCPClientSocketChannel, self).__init__(s)


class TCPServerSocketChannel(SocketChannel):
    """
    Convenience subclass of :class:`SocketChannel` that waits for a remote
    client to connect.

    Args:
        host(str): The hostname or IP address to connect to. Defaults to all
            IP adresses.
        port(int): The port number to connect to. Defaults to a random port
            chosen by the OS.
    """

    def __init__(self, host='', port=0):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((host, port))
        s.listen(1)
        cs, _ = s.accept()
        s.close()
        super(TCPServerSocketChannel, self).__init__(cs)


class SSHClient(paramiko.client.SSHClient):
    """
    Wrapper around Paramiko's SSHClient that produces Flow instances when
    calling :meth:`SSHClient.execute` and :meth:`SSHClient.invoke_shell`.

    Since `pwnypack` is usually used for CTFs, the missing host key policy
    is set to warn and accept.
    """

    def __init__(self):
        super(SSHClient, self).__init__()
        self.set_missing_host_key_policy(paramiko.client.WarningPolicy())

    def execute(self, command, pty=False, echo=False):
        """
        Execute `command` on the server this :class:`SSHClient` instance is connected
        to.

        Args:
            command(str): The command to execute on the remote server.
            pty(bool): Request a pseudo-terminal from the server. Note: If
                this is `False` (the default) then `stderr` will be combined
                into `stdout` to prevent the buffers from filling up.
            echo(bool): Whether to write the read data to stdout.

        Returns:
            :class:`Flow`: A Flow instance initialised with the SSH channel.
        """

        channel = self.get_transport().open_session()
        if pty:
            channel.get_pty()
        else:
            channel.set_combine_stderr(True)
        channel.exec_command(command)
        return Flow(SocketChannel(channel), echo=echo)

    def invoke_shell(self, pty=True, echo=False):
        """
        Start a new shell on the server this :class:`SSHClient` is connected
        to.

        Args:
            pty(bool): Request a pseudo-terminal from the server. Note: If
                this is `False` then `stderr` will be combined into `stdout`
                to prevent the buffers from filling up.
            echo(bool): Whether to write the read data to stdout.

        Returns:
            :class:`Flow`: A Flow instance initialised with the SSH channel.
        """

        channel = self.get_transport().open_session()
        if pty:
            channel.get_pty()
        else:
            channel.set_combine_stderr(True)
        channel.invoke_shell()
        return Flow(SocketChannel(channel), echo=echo)


class Flow(object):
    """
    The core class of *Flow*. Takes a channel and exposes synchronous
    utility functions for communications.

    Usually, you'll use the convenience classmethods :meth:`connect_tcp`
    or :meth:`execute` instead of manually creating the constructor
    directly.

    Args:
        channel(``Channel``): A channel.
        echo(bool): Whether or not to echo all input / output.
    """

    def __init__(self, channel, echo=False):
        self.channel = channel
        self.echo = echo

    def read(self, n, echo=None):
        """
        Read *n* bytes from the channel.

        Args:
            n(int): The number of bytes to read from the channel.
            echo(bool): Whether to write the read data to stdout.

        Returns:
            bytes: *n* bytes of data.

        Raises:
            EOFError: If the channel was closed.
        """

        d = self.channel.read(n)
        if echo or (echo is None and self.echo):
            sys.stdout.write(d.decode('latin1'))
            sys.stdout.flush()
        return d

    def read_eof(self, echo=None):
        """
        Read until the channel is closed.

        Args:
            echo(bool): Whether to write the read data to stdout.

        Returns:
            bytes: The read data.
        """

        d = b''
        while True:
            try:
                d += self.read(1, echo)
            except EOFError:
                return d

    def read_until(self, s, echo=None):
        """
        Read until a certain string is encountered..

        Args:
            s(bytes): The string to wait for.
            echo(bool): Whether to write the read data to stdout.

        Returns:
            bytes: The data up to and including *s*.

        Raises:
            EOFError: If the channel was closed.
        """

        s_len = len(s)
        buf = self.read(s_len, echo)

        while buf[-s_len:] != s:
            buf += self.read(1, echo)

        return buf

    until = read_until  #: Alias of :meth:`read_until`.

    def readlines(self, n, echo=None):
        """
        Read *n* lines from channel.

        Args:
            n(int): The number of lines to read.
            echo(bool): Whether to write the read data to stdout.

        Returns:
            list of bytes: *n* lines which include new line characters.

        Raises:
            EOFError: If the channel was closed before *n* lines were read.
        """

        return [
            self.until(b'\n', echo)
            for _ in range(n)
        ]

    def readline(self, echo=None):
        """
        Read 1 line from channel.

        Args:
            echo(bool): Whether to write the read data to stdout.

        Returns:
            bytes: The read line which includes new line character.

        Raises:
            EOFError: If the channel was closed before a line was read.
        """

        return self.readlines(1, echo)[0]

    def write(self, data, echo=None):
        """
        Write data to channel.

        Args:
            data(bytes): The data to write to the channel.
            echo(bool): Whether to echo the written data to stdout.

        Raises:
            EOFError: If the channel was closed before all data was sent.
        """

        if echo or (echo is None and self.echo):
            sys.stdout.write(data.decode('latin1'))
            sys.stdout.flush()
        self.channel.write(data)

    def writelines(self, lines, sep=b'\n', echo=None):
        """
        Write a list of byte sequences to the channel and terminate them
        with a separator (line feed).

        Args:
            lines(list of bytes): The lines to send.
            sep(bytes): The separator to use after each line.
            echo(bool): Whether to echo the written data to stdout.

        Raises:
            EOFError: If the channel was closed before all data was sent.
        """

        self.write(sep.join(lines + [b'']), echo)

    def writeline(self, line=b'', sep=b'\n', echo=None):
        """
        Write a byte sequences to the channel and terminate it with carriage
        return and line feed.

        Args:
            line(bytes): The line to send.
            sep(bytes): The separator to use after each line.
            echo(bool): Whether to echo the written data to stdout.

        Raises:
            EOFError: If the channel was closed before all data was sent.
        """

        self.writelines([line], sep, echo)

    def close(self):
        """
        Gracefully close the channel.
        """

        self.channel.close()

    def kill(self):
        """
        Terminate the channel immediately.
        """

        self.channel.kill()

    def interact(self):
        """
        Interact with the socket. This will send all keyboard input to the
        socket and input from the socket to the console until an EOF occurs.
        """

        sockets = [sys.stdin, self.channel]
        while True:
            ready = select.select(sockets, [], [])[0]

            if sys.stdin in ready:
                line = sys.stdin.readline().encode('utf-8')
                if not line:
                    break
                self.write(line)

            if self.channel in ready:
                self.read(1, echo=True)

    @classmethod
    def execute(cls, executable, *arguments, **kwargs):
        """execute(executable, argument..., redirect_stderr=False, echo=False):

        Set up a :class:`ProcessChannel` and create a :class:`Flow` instance
        for it.

        Args:
            executable(str): The executable to start.
            argument...(list of str): The arguments to pass to the executable.
            redirect_stderr(bool): Whether to also capture the output of stderr.
            echo(bool): Whether to echo read/written data to stdout by default.

        Returns:
            :class:`Flow`: A Flow instance initialised with the process
                channel.
        """

        echo = kwargs.pop('echo', False)
        return cls(ProcessChannel(executable, *arguments, **kwargs), echo=echo)

    @classmethod
    def connect_tcp(cls, host, port, echo=False):
        """
        Set up a :class:`TCPClientSocketChannel` and create a :class:`Flow`
        instance for it.

        Args:
            host(str): The hostname or IP address to connect to.
            port(int): The port number to connect to.
            echo(bool): Whether to echo read/written data to stdout by default.

        Returns:
            :class:`Flow`: A Flow instance initialised with the TCP socket
                channel.
        """

        return cls(TCPClientSocketChannel(host, port), echo=echo)

    @classmethod
    def listen_tcp(cls, host='', port=0, echo=False):
        """
        Set up a :class:`TCPServerSocketChannel` and create a :class:`Flow`
        instance for it.

        Args:
            host(str): The hostname or IP address to bind to.
            port(int): The port number to listen on.
            echo(bool): Whether to echo read/written data to stdout by default.

        Returns:
            :class:`Flow`: A Flow instance initialised with the TCP socket
                channel.
        """

        return cls(TCPServerSocketChannel(host, port), echo=echo)

    @staticmethod
    def connect_ssh(*args, **kwargs):
        """
        Create a new connected :class:`SSHClient` instance. All arguments
        are passed to :meth:`SSHClient.connect`.
        """

        client = SSHClient()
        client.connect(*args, **kwargs)
        return client

    @classmethod
    def execute_ssh(cls, command, *args, **kwargs):
        """execute_ssh(command, arguments..., pty=False, echo=False)

        Execute `command` on a remote server. It first calls
        :meth:`Flow.connect_ssh` using all positional and keyword
        arguments, then calls :meth:`SSHClient.execute` with the command
        and pty / echo options.

        Args:
            command(str): The command to execute on the remote server.
            arguments...: The options for the SSH connection.
            pty(bool): Request a pseudo-terminal from the server.
            echo(bool): Whether to echo read/written data to stdout by default.

        Returns:
            :class:`Flow`: A Flow instance initialised with the SSH channel.
        """

        pty = kwargs.pop('pty', False)
        echo = kwargs.pop('echo', False)
        client = cls.connect_ssh(*args, **kwargs)
        f = client.execute(command, pty=pty, echo=echo)
        f.client = client
        return f


    @classmethod
    def invoke_ssh_shell(cls, *args, **kwargs):
        """invoke_ssh(arguments..., pty=False, echo=False)

        Star a new shell on a remote server. It first calls
        :meth:`Flow.connect_ssh` using all positional and keyword
        arguments, then calls :meth:`SSHClient.invoke_shell` with the
        pty / echo options.

        Args:
            arguments...: The options for the SSH connection.
            pty(bool): Request a pseudo-terminal from the server.
            echo(bool): Whether to echo read/written data to stdout by default.

        Returns:
            :class:`Flow`: A Flow instance initialised with the SSH channel.
        """

        pty = kwargs.pop('pty', True)
        echo = kwargs.pop('echo', False)
        client = cls.connect_ssh(*args, **kwargs)
        f = client.invoke_shell(pty=pty, echo=echo)
        f.client = client
        return f



import pwny


def test_cycle():
    assert pwny.cycle(64) == 'AAAABAAACAAADAAAEAAAFAAAGAAAHAAAIAAAJAAAKAAALAAAMAAANAAAOAAAPAAA'


def test_cycle_width():
    assert pwny.cycle(64, width=2) == 'AABACADAEAFAGAHAIAJAKALAMANAOAPAQARASATAUAVAWAXAYAZBBCBDBEBFBGBH'


def test_cycle_find():
    assert pwny.cycle_find('PAAA') == 60


def test_cycle_find_start():
    assert pwny.cycle_find('AAAA') == 0


def test_cycle_find_not_found():
    assert pwny.cycle_find('\x00', width=1) == -1



"""
This module contains functions that allow you to manipulate, encode or decode
strings and byte sequences.
"""


import base64
import re
import string
import six
import codecs
import pwnypack.main
import binascii
from six.moves import range
from six.moves.urllib.parse import quote, quote_plus, unquote, unquote_plus, urlencode, parse_qs
try:
    from collections import Counter
except ImportError:
    from counter import Counter


__all__ = [
    'xor',
    'rot13',
    'caesar',
    'enhex',
    'dehex',
    'enb64',
    'deb64',
    'enurlform',
    'deurlform',
    'enurlquote',
    'deurlquote',
    'frequency',
]


def xor(key, data):
    """
    Perform cyclical exclusive or operations on ``data``.

    The ``key`` can be a an integer *(0 <= key < 256)* or a byte sequence. If
    the key is smaller than the provided ``data``, the ``key`` will be
    repeated.

    Args:
        key(int or bytes): The key to xor ``data`` with.
        data(bytes): The data to perform the xor operation on.

    Returns:
        bytes: The result of the exclusive or operation.

    Examples:
        >>> from pwny import *
        >>> xor(5, b'ABCD')
        b'DGFA'
        >>> xor(5, b'DGFA')
        b'ABCD'
        >>> xor(b'pwny', b'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
        b'15-=51)19=%5=9!)!%=-%!9!)-'
        >>> xor(b'pwny', b'15-=51)19=%5=9!)!%=-%!9!)-')
        b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    """

    if type(key) is int:
        key = six.int2byte(key)
    key_len = len(key)

    return b''.join(
        six.int2byte(c ^ six.indexbytes(key, i % key_len))
        for i, c in enumerate(six.iterbytes(data))
    )


def caesar(shift, data, shift_ranges=('az', 'AZ')):
    """
    Apply a caesar cipher to a string.

    The caesar cipher is a substition cipher where each letter in the given
    alphabet is replaced by a letter some fixed number down the alphabet.

    If ``shift`` is ``1``, *A* will become *B*, *B* will become *C*, etc...

    You can define the alphabets that will be shift by specifying one or more
    shift ranges. The characters will than be shifted within the given ranges.

    Args:
        shift(int): The shift to apply.
        data(str): The string to apply the cipher to.
        shift_ranges(list of str): Which alphabets to shift.

    Returns:
        str: The string with the caesar cipher applied.

    Examples:
        >>> caesar(16, 'Pwnypack')
        'Fmdofqsa'
        >>> caesar(-16, 'Fmdofqsa')
        'Pwnypack'
        >>> caesar(16, 'PWNYpack', shift_ranges=('AZ',))
        'FMDOpack'
        >>> caesar(16, 'PWNYpack', shift_ranges=('Az',))
        '`g^iFqsA'
    """
    alphabet = dict(
        (chr(c), chr((c - s + shift) % (e - s + 1) + s))
        for s, e in map(lambda r: (ord(r[0]), ord(r[-1])), shift_ranges)
        for c in range(s, e + 1)
    )
    return ''.join(alphabet.get(c, c) for c in data)


rot13_encode = codecs.getencoder('rot-13')
rot13 = lambda d: rot13_encode(d)[0]
rot13.__doc__ = """
Rotate all characters in the alphabets A-Z and a-z by 13 positions in the
alphabet. This is a :func:`caesar` shift of 13 along the fixed alphabets
``A-Z`` and ``a-z``.

Args:
    d(str): The string to the apply the cipher to.

Returns:
    str: The string with the rot13 cipher applied.

Examples:
    >>> rot13('whax')
    'junk'
    >>> rot13('junk')
    'whax'
"""


def enhex(d, separator=''):
    """
    Convert bytes to their hexadecimal representation, optionally joined by a
    given separator.

    Args:
        d(bytes): The data to convert to hexadecimal representation.
        separator(str): The separator to insert between hexadecimal tuples.

    Returns:
        str: The hexadecimal representation of ``d``.

    Examples:
        >>> from pwny import *
        >>> enhex(b'pwnypack')
        '70776e797061636b'
        >>> enhex(b'pwnypack', separator=' ')
        '70 77 6e 79 70 61 63 6b'
    """

    v = binascii.hexlify(d).decode('ascii')
    if separator:
        return separator.join(
            v[i:i+2]
            for i in range(0, len(v), 2)
        )
    else:
        return v


dehex_clean = re.compile('[^a-fA-F0-9]')
dehex = lambda d: binascii.unhexlify(dehex_clean.sub('', d).encode('ascii'))
dehex.__doc__ = """
Convert a hexadecimal representation of a byte sequence to bytes. All
non-hexadecimal characters will be removed from the input.

Args:
    d(str): The string of hexadecimal tuples.

Returns:
    bytes: The byte sequence represented by ``d``.

Examples:
    >>> from pwny import *
    >>> dehex('70776e797061636b')
    b'pwnypack'
    >>> dhex('70 77 6e 79 70 61 63 6b')
    b'pwnypack'
"""


enb64 = lambda d: base64.b64encode(d).decode('ascii')
enb64.__doc__ = """
Convert bytes to their base64 representation.

Args:
    d(bytes): The data to convert to its base64 representation.

Returns:
    str: The base64 representation of ``d``.

Example:
    >>> from pwny import *
    >>> enb64(b'pwnypack')
    'cHdueXBhY2s='
"""


deb64 = lambda d: base64.b64decode(d.encode('ascii'))
deb64.__doc__ = """
Convert a base64 representation back to its original bytes.

Args:
    d(str): The base64 representation to decode.

Returns:
    bytes: The bytes represented by ``d``.

Example:
    >>> from pwny import *
    >>> deb64('cHdueXBhY2s=')
    b'pwnypack'
"""


def enurlform(q):
    """
    Convert a dictionary to a URL encoded query string.

    Args:
        q(dict): The query to encode.

    Returns:
        str: The urlencoded representation of ``q``.

    Example:
        >>> from pwny import *
        >>> enurlform({'foo': 'bar', 'baz': ['quux', 'corge']})
        'foo=bar&baz=quux&baz=corge'
    """
    return urlencode(q, doseq=True)


def deurlform(d):
    """
    Convert a URL encoded query string to a dictionary.

    Args:
        d(str): The URL encoded query string.

    Returns:
        dict: A dictionary containing each key and all its values as a list.

    Example:
        >>> from pwny import *
        >>> deurlform('foo=bar&baz=quux&baz=corge')
        {'foo': ['bar'], 'baz': ['quux', 'corge']}
    """
    return parse_qs(d)


def enurlquote(v, plus=False):
    """
    Percent encode a string for use in an URL.

    Args:
        v(str): The value to percent encode.
        plus(bool): Use a plus symbol for spaces, otherwise use %20.

    Returns:
        str: The percent encoded string.

    Example:
        >>> from pwny import *
        >>> enurlquote('Foo Bar/Baz', True)
        'Foo+Bar/Baz
    """
    return quote_plus(v) if plus else quote(v)


def deurlquote(d, plus=False):
    """
    Decode a percent encoded string.

    Args:
        d(str): The percent encoded value to decode.
        plus(bool): Parse a plus symbol as a space.

    Returns:
        str: The decoded version of the percent encoded of ``d``.

    Example:
        >>> from pwny import *
        >>> deurlquote('Foo+Bar/Baz')
        'Foo Bar/Baz'
    """
    return unquote_plus(d) if plus else unquote(d)


frequency = lambda v: dict(Counter(v))
frequency.__doc__ = """
Perform a frequency analysis on a byte sequence or string.

Args:
    d(bytes or str): The sequence to analyse.

Returns:
    dict: A dictionary of unique elements in ``d`` and how often the occur.

Example:
    >>> frequency('pwnypack')
    {'a': 1, 'c': 1, 'k': 1, 'n': 1, 'p': 2, 'w': 1, 'y': 1}
"""


@pwnypack.main.register('xor')
def xor_app(parser, cmd, args):  # pragma: no cover
    """
    Xor a value with a key.
    """

    parser.add_argument(
        '-d', '--dec',
        help='interpret the key as a decimal integer',
        dest='type',
        action='store_const',
        const=int
    )
    parser.add_argument(
        '-x', '--hex',
        help='interpret the key as an hexadecimal integer',
        dest='type',
        action='store_const',
        const=lambda v: int(v, 16)
    )
    parser.add_argument('key', help='the key to xor the value with')
    parser.add_argument('value', help='the value to xor, read from stdin if omitted', nargs='?')

    args = parser.parse_args(args)
    if args.type is not None:
        args.key = args.type(args.key)

    return xor(args.key, pwnypack.main.binary_value_or_stdin(args.value))


@pwnypack.main.register('caesar')
def caesar_app(parser, cmd, args):  # pragma: no cover
    """
    Caesar crypt a value with a key.
    """

    parser.add_argument('shift', type=int, help='the shift to apply')
    parser.add_argument('value', help='the value to caesar crypt, read from stdin if omitted', nargs='?')
    parser.add_argument(
        '-s', '--shift-range',
        dest='shift_ranges',
        action='append',
        help='specify a character range to shift (defaults to a-z, A-Z)'
    )

    args = parser.parse_args(args)
    if not args.shift_ranges:
        args.shift_ranges = ['az', 'AZ']

    return caesar(args.shift, pwnypack.main.string_value_or_stdin(args.value), args.shift_ranges)


@pwnypack.main.register('rot13')
def rot13_app(parser, cmd, args):  # pragma: no cover
    """
    rot13 encrypt a value.
    """

    parser.add_argument('value', help='the value to rot13, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return rot13(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('enb64')
def enb64_app(parser, cmd, args):  # pragma: no cover
    """
    base64 encode a value.
    """

    parser.add_argument('value', help='the value to base64 encode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return enb64(pwnypack.main.binary_value_or_stdin(args.value))


@pwnypack.main.register('deb64')
def deb64_app(parser, cmd, args):  # pragma: no cover
    """
    base64 decode a value.
    """

    parser.add_argument('value', help='the value to base64 decode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return deb64(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('enhex')
def enhex_app(parser, cmd, args):  # pragma: no cover
    """
    hex encode a value.
    """

    parser.add_argument('value', help='the value to hex encode, read from stdin if omitted', nargs='?')
    parser.add_argument(
        '--separator', '-s',
        default='',
        help='the separator to place between hex tuples'
    )
    args = parser.parse_args(args)
    return enhex(pwnypack.main.binary_value_or_stdin(args.value), args.separator)


@pwnypack.main.register('dehex')
def dehex_app(parser, cmd, args):  # pragma: no cover
    """
    hex decode a value.
    """

    parser.add_argument('value', help='the value to base64 decode, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    return dehex(pwnypack.main.string_value_or_stdin(args.value))


@pwnypack.main.register('enurlform')
def enurlform_app(parser, cmd, args):  # pragma: no cover
    """
    encode a series of key=value pairs into a query string.
    """

    parser.add_argument('values', help='the key=value pairs to URL encode', nargs='+')
    args = parser.parse_args(args)
    return enurlform(dict(v.split('=', 1) for v in args.values))


@pwnypack.main.register('deurlform')
def deurlform_app(parser, cmd, args):  # pragma: no cover
    """
    decode a query string into its key value pairs.
    """

    parser.add_argument('value', help='the query string to decode')
    args = parser.parse_args(args)
    return ' '.join('%s=%s' % (key, value) for key, values in deurlform(args.value).items() for value in values)


@pwnypack.main.register('frequency')
def frequency_app(parser, cmd, args):  # pragma: no cover
    """
    perform frequency analysis on a value.
    """

    parser.add_argument('value', help='the value to analyse, read from stdin if omitted', nargs='?')
    args = parser.parse_args(args)
    data = frequency(six.iterbytes(pwnypack.main.binary_value_or_stdin(args.value)))
    return '\n'.join(
        '0x%02x (%c): %d' % (key, chr(key), value)
        if key >= 32 and chr(key) in string.printable else
        '0x%02x ---: %d' % (key, value)
        for key, value in data.items()
    )



"""
The fmtstring module allows you to build format strings that can be used to
exploit format string bugs (`printf(buf);`).
"""


import pwnypack.target
import pwnypack.packing


__all__ = ['fmtstring']


FMTSTRING_OPS = {
    1: b'hhn',
    2: b'hn',
    4: b'n',
}


def fmtstring(offset, writes, written=0, max_width=2, target=None):
    """
    Build a format string that writes given data to given locations. Can be
    used easily create format strings to exploit format string bugs.

    `writes` is a list of 2- or 3-item tuples. Each tuple represents a memory
    write starting with an absolute address, then the data to write as an
    integer and finally the width (1, 2, 4 or 8) of the write.

    :func:`fmtstring` will break up the writes and try to optimise the order
    to minimise the amount of dummy output generated.

    Args:
        offset(int): The parameter offset where the format string start.
        writes(list): A list of 2 or 3 item tuples.
        written(int): How many bytes have already been written before the
            built format string starts.
        max_width(int): The maximum width of the writes (1, 2 or 4).
        target(:class:`pwnypack.target.Target`): The target architecture.

    Returns:
        bytes: The format string that will execute the specified memory
            writes.

    Example:
        The following example will (on a 32bit architecture) build a format
        string that write 0xc0debabe to the address 0xdeadbeef and the byte
        0x90 to 0xdeadbeef + 4 assuming that the input buffer is located at
        offset 3 on the stack.

        >>> from pwny import *
        >>> fmtstring(3, [(0xdeadbeef, 0xc0debabe), (0xdeadbeef + 4, 0x90, 1)])
    """

    if max_width not in (1, 2, 4):
        raise ValueError('max_width should be 1, 2 or 4')

    if target is None:
        target = pwnypack.target.target

    addrs = []
    cmds = []
    piece_writes = []

    for write in writes:
        if len(write) == 2:
            addr, value = write
            width = target.bits // 8
        else:
            addr, value, width = write
            if width not in (1, 2, 4, 8):
                raise ValueError('Invalid write width')

        piece_width = min(max_width, width)
        piece_value = getattr(pwnypack.packing, 'P%d' % (8 * width))(value, target=target)
        piece_unpack = getattr(pwnypack.packing, 'U%d' % (piece_width * 8))

        for i in range(0, width, piece_width):
            piece_writes.append((piece_width, addr, piece_unpack(piece_value[i:i + piece_width], target=target)))
            addr += piece_width

    written += len(piece_writes) * int(target.bits) // 8

    piece_writes.sort(key=lambda w_a_v: (w_a_v[2] - written) % (2 ** (max_width * 8)))

    for piece_width, piece_addr, piece_value in piece_writes:
        addrs.append(pwnypack.packing.P(piece_addr, target=target))

        piece_modulo = 2 ** (piece_width * 8)

        padding = (piece_value - written) % piece_modulo
        if padding:
            cmds.append(b'%' + str(padding).encode('ascii') + b'c')
        written = piece_value

        cmds.append(b'%' + str(offset).encode('ascii') + b'$' + FMTSTRING_OPS[piece_width])
        offset += 1

    return b''.join(addrs + cmds)



from pwnypack.target import *
from pwnypack.packing import *
from pwnypack.util import *
from pwnypack.flow import *
from pwnypack.asm import *
from pwnypack.elf import *
from pwnypack.codec import *
from pwnypack.shell import *
from pwnypack.rop import *
from pwnypack.fmtstring import *
from pwnypack.php import *



import six


__all__ = [
    'php_serialize',
    'PhpObject',
]


def php_serialize(value):
    """
    Serialize a value for use with PHP's deserialize() function. This function
    can serialize bytes, strings, integers, floats, booleans, None, lists,
    dicts and custom objects implementing __php__().

    Args:
        value: The value to serialize.

    Returns:
        bytes: The serialized form of `value` ready to be unserialized by PHP.

    Example:
        >>> from pwny import *
        >>> php_serialize([b'foo', u'bar', 42, 2.5, True, None, {'a': 'b'}])
        b'a:7:{i:0;s:3:"foo";i:1;s:3:"bar";i:2;i:42;i:3;d:2.5;i:4;b:1;i:5;N;i:6;a:1:{s:1:"a";s:1:"b";}}'
    """

    def serialize_array(items):
        content = b''.join(
            php_serialize(i) + php_serialize(v)
            for i, v in items
        )
        return 'a:{0}:'.format(len(value)).encode('utf8') + b'{' + content + b'}'

    def serialize_str(prefix, item):
        return prefix + b':' + str(item).encode('utf8') + b';'

    if isinstance(value, six.binary_type):
        return b's:' + str(len(value)).encode('utf8') + b':"' + value + b'";'
    elif isinstance(value, six.text_type):
        return php_serialize(value.encode('utf8'))
    elif isinstance(value, bool):
        return serialize_str(b'b', 1 if value else 0)
    elif isinstance(value, int):
        return serialize_str(b'i', value)
    elif isinstance(value, float):
        return serialize_str(b'd', value)
    elif value is None:
        return b'N;'
    elif isinstance(value, (list, tuple)):
        return serialize_array(enumerate(value))
    elif isinstance(value, dict):
        return serialize_array(six.iteritems(value))
    else:
        return value.__php__()


class PhpObject(object):
    """
    Helper class to represent PHP objects for serialization using
    :func:`php_serialize`.

    Instances of this class act like a dictionary of properties that should
    be set on the deserialized PHP instance. You can prefix the property
    names with ``'public '``,  ``'protected '`` or ``'private '`` to ensure
    the correct instance variables are set.

    Arguments:
        class_name(str): The name of the PHP class to use when
            deserializing.
        properties(dict): The properties to deserialize in this instance.

    Example:
        >>> from pwny import *
        >>> o = PhpObject('Foo\\Bar', {'protected fg': '#000000'})
        >>> php_serialize(o)
        b'O:7:"Foo\\Bar":1:{s:5:"\\x00*\\x00fg";s:7:"#000000";}'
    """

    def __init__(self, class_name, properties=None):
        self.class_name = class_name
        self.items = {}
        for k, v in six.iteritems(properties or {}):
            self[k] = v

    def _mogrify_key(self, key):
        if key.startswith('protected '):
            return '\0*\0' + key.split(' ', 1)[1]
        elif key.startswith('private '):
            return self.class_name + key.split(' ', 1)[1]
        elif key.startswith('public '):
            return key.split(' ', 1)[1]
        else:
            return key

    def __setitem__(self, key, value):
        self.items[self._mogrify_key(key)] = value

    def __getitem__(self, key):
        return self.items[self._mogrify_key(key)]

    def __php__(self):
        properties = b''.join(
            php_serialize(k) + php_serialize(v)
            for k, v in six.iteritems(self.items)
        )
        return 'O:{0}:"{1}":{2}:'.format(len(self.class_name), self.class_name, len(self.items)).encode('utf8') + \
            b'{' + properties + b'}'



from setuptools import setup
import sys
import os


__version__ = '0.7.1'


requires = [
    'six',
    'nose',
    'capstone',
    'paramiko',
]


if sys.version_info[:2] < (2, 7):
    requires.extend([
        'counter',
        'ordereddict',
        'argparse',
    ])


if sys.version_info[:2] < (3, 4):
    requires.append('enum34')


def read_file(filename):
    try:
        with open(os.path.join(os.path.dirname(__file__), filename)) as f:
            return f.read()
    except IOError:
        return ''


setup(
    name='pwnypack',
    packages=['pwny', 'pwnypack'],
    version=__version__,
    description='Official Certified Edible Dinosaurs CTF toolkit.',
    long_description=read_file('README.rst') + '\n' + read_file('changelog.rst'),
    author='Ingmar Steen',
    author_email='iksteen@gmail.com',
    url='https://github.com/edibledinos/pwnypack/',
    download_url='https://github.com/edibledinos/pwnypack/tarball/v%s' % __version__,
    install_requires=requires,
    tests_require=['mock', 'coverage'],
    entry_points={
        'console_scripts': [
            'pwny=pwnypack.main:main',
        ],
    },
    keywords=['wargame', 'ctf'],
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
    ],
)



"""
The :class:`Target` class describes the architecture of a targeted machine,
executable or environment. It encodes the generic architecture, the word
size, the byte order and an architecture dependant mode.

It is used throughout *pwnypack* to determine how data should be interpreted
or emitted.
"""


import platform
from enum import IntEnum, Enum
import sys


__all__ = [
    'Target',
    'target',
]


class Target(object):
    class Arch(Enum):
        """
        Describes the general architecture of a target.
        """

        x86 = 'x86'          #: X86 architecture.
        arm = 'arm'          #: ARM architecture.
        unknown = 'unknown'  #: Any other architecture.

    class Bits(IntEnum):
        """
        The target architecture's word size.
        """

        bits_32 = 32  #: 32 bit word size.
        bits_64 = 64  #: 64 bit word size.

    class Endian(IntEnum):
        """
        The target architecture's byte order.
        """

        little = 0  #: Little endian.
        big = 1     #: Big endian.

    class Mode(IntEnum):
        """
        Architecture dependant mode flags.
        """

        arm_v8 = 1 << 0       #: Use ARM V8 instruction set
        arm_thumb = 1 << 1    #: Use ARM Thumb instruction set
        arm_m_class = 1 << 2  #: Use ARMv7-M instruction set

    _DEFAULT_ARCH = {
        'i386': Arch.x86,
        'x86_64': Arch.x86,
    }

    _VALID_BITS = {}

    _DEFAULT_BITS = {
        Arch.x86: Bits.bits_32,
        Arch.arm: Bits.bits_32,
    }

    _VALID_ENDIAN = {
        Arch.x86: [Endian.little],
    }

    _DEFAULT_ENDIAN = {
        Arch.x86: Endian.little,
        Arch.arm: Endian.little,
    }

    _arch = None
    _bits = None
    _endian = None
    _mode = None

    def __init__(self, arch=None, bits=None, endian=None, mode=0):
        if arch is None:
            arch = self._DEFAULT_ARCH.get(platform.machine(), Target.Arch.unknown)

            if bits is None:
                bits = Target.Bits(64 if platform.architecture()[0] == '64bit' else 32)

            if endian is None:
                endian = Target.Endian.__members__[sys.byteorder]

        self.arch = arch
        self.bits = bits
        self.endian = endian
        self.mode = mode

    @property
    def arch(self):
        """
        The target's architecture. One of :class:`Target.Arch`.
        """
        return self._arch

    @arch.setter
    def arch(self, arch):
        if arch is None:
            arch = Target.Arch.unknown
        self._arch = Target.Arch(arch)

    @property
    def bits(self):
        """
        The target architecture word size. One of :class:`Target.Bits`.
        """
        if self._bits is None:
            value = self._DEFAULT_BITS.get(self.arch)
            if value is None:
                raise NotImplementedError('Could not determine the default word size of %s architecture.' % self.arch)
            return value
        else:
            return self._bits

    @bits.setter
    def bits(self, value):
        if value is None:
            self._bits = None
        else:
            self._bits = Target.Bits(value)
            assert self._bits in self._VALID_BITS.get(
                self.arch,
                [Target.Bits.bits_32, Target.Bits.bits_64]
            ), '%s not supported on %s' % (self._bits, self._arch)

    @property
    def endian(self):
        """
        The target architecture byte order. One of :class:`Target.Endian`.
        """
        if self._endian is None:
            value = self._DEFAULT_ENDIAN[self.arch]
            if value is None:
                raise NotImplementedError('Could not determine the default byte order of %s architecture.' % self.arch)
            return value
        else:
            return self._endian

    @endian.setter
    def endian(self, endian):
        if endian is None:
            self._endian = None
        else:
            self._endian = Target.Endian(endian)
            assert self._endian in self._VALID_ENDIAN.get(
                self.arch,
                [Target.Endian.little, Target.Endian.big]
            ), '%s not supported on %s' % (self._endian, self._arch)

    @property
    def mode(self):
        """
        The target architecture dependant flags. OR'ed values of :class:`Target.Mode`.
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = int(value)

    def assume(self, other):
        """
        Assume the identity of another target. This can be useful to make the
        global target assume the identity of an ELF executable.

        Arguments:
            other(:class:`Target`): The target whose identity to assume.

        Example:
            >>> from pwny import *
            >>> target.assume(ELF('my-executable'))
        """

        self._arch = other._arch
        self._bits = other._bits
        self._endian = other._endian
        self._mode = other._mode


target = Target()
"""
The global, default target. If no targeting information is provided to a
function, this is the target that will be used.
"""



from nose.tools import raises
import pwny


target_little_endian = pwny.Target(arch=pwny.Target.Arch.unknown, endian=pwny.Target.Endian.little)
target_big_endian = pwny.Target(arch=pwny.Target.Arch.unknown, endian=pwny.Target.Endian.big)


def test_pack():
    assert pwny.pack('I', 0x41424344) == b'DCBA'


def test_pack_format_with_endian():
    assert pwny.pack('>I', 0x41424344) == b'ABCD'


def test_pack_explicit_endian():
    assert pwny.pack('I', 0x41424344, endian=pwny.Target.Endian.big) == b'ABCD'


def test_pack_explicit_target():
    assert pwny.pack('I', 0x41424344, target=target_big_endian) == b'ABCD'


@raises(NotImplementedError)
def test_pack_invalid_endian():
    pwny.pack('I', 1, endian='invalid')


def test_unpack():
    assert pwny.unpack('I', b'DCBA') == (0x41424344,)


def test_unpack_format_with_endian():
    assert pwny.unpack('>I', b'ABCD') == (0x41424344,)


def test_unpack_explicit_endian():
    assert pwny.unpack('I', b'ABCD', endian=pwny.Target.Endian.big) == (0x41424344,)


def test_unpack_explicit_target():
    assert pwny.unpack('I', b'ABCD', target=target_big_endian) == (0x41424344,)


@raises(NotImplementedError)
def test_unpack_invalid_endian():
    pwny.unpack('I', 'AAAA', endian='invalid')


def test_pack_size():
    # This tests both pack_size in general as well as not padding the byte.
    assert pwny.pack_size('bq') == 9


short_signed_data = [
    [8, -0x7f, b'\x81'],
    [16, -0x7fff, b'\x80\x01'],
    [32, -0x7fffffff, b'\x80\x00\x00\x01'],
    [64, -0x7fffffffffffffff, b'\x80\x00\x00\x00\x00\x00\x00\x01'],
]


short_unsigned_data = [
    [8, 0x61, b'a'],
    [16, 0x6162, b'ab'],
    [32, 0x61626364, b'abcd'],
    [64, 0x6162636465666768, b'abcdefgh'],
]


def test_short_form_pack():
    for width, num, bytestr in short_signed_data:
        f = 'p%d' % width
        yield check_short_form_pack, f, num, bytestr[::-1]
        yield check_short_form_pack_endian, f, num, bytestr[::-1], pwny.Target.Endian.little
        yield check_short_form_pack_endian, f, num, bytestr, pwny.Target.Endian.big

    for width, num, bytestr in short_unsigned_data:
        f = 'P%d' % width
        yield check_short_form_pack, f, num, bytestr[::-1]
        yield check_short_form_pack_endian, f, num, bytestr[::-1], pwny.Target.Endian.little
        yield check_short_form_pack_endian, f, num, bytestr, pwny.Target.Endian.big


def test_short_form_unpack():
    for width, num, bytestr in short_signed_data:
        f = 'u%d' % width
        yield check_short_form_unpack, f, num, bytestr[::-1]
        yield check_short_form_unpack_endian, f, num, bytestr[::-1], pwny.Target.Endian.little
        yield check_short_form_unpack_endian, f, num, bytestr, pwny.Target.Endian.big

    for width, num, bytestr in short_unsigned_data:
        f = 'U%d' % width
        yield check_short_form_unpack, f, num, bytestr[::-1]
        yield check_short_form_unpack_endian, f, num, bytestr[::-1], pwny.Target.Endian.little
        yield check_short_form_unpack_endian, f, num, bytestr, pwny.Target.Endian.big


def test_pointer_pack():
    yield check_short_form_pack, 'p', -66052, b'\xfc\xfd\xfe\xff'
    yield check_short_form_pack_endian, 'p', -66052, b'\xfc\xfd\xfe\xff', pwny.Target.Endian.little
    yield check_short_form_pack_endian, 'p', -66052, b'\xff\xfe\xfd\xfc', pwny.Target.Endian.big

    yield check_short_form_pack, 'P', 4294901244, b'\xfc\xfd\xfe\xff'
    yield check_short_form_pack_endian, 'P', 4294901244, b'\xfc\xfd\xfe\xff', pwny.Target.Endian.little
    yield check_short_form_pack_endian, 'P', 4294901244, b'\xff\xfe\xfd\xfc', pwny.Target.Endian.big


def test_pointer_unpack():
    yield check_short_form_unpack, 'u', -66052, b'\xfc\xfd\xfe\xff'
    yield check_short_form_unpack_endian, 'u', -66052, b'\xfc\xfd\xfe\xff', pwny.Target.Endian.little
    yield check_short_form_unpack_endian, 'u', -66052, b'\xff\xfe\xfd\xfc', pwny.Target.Endian.big

    yield check_short_form_unpack, 'U', 4294901244, b'\xfc\xfd\xfe\xff'
    yield check_short_form_unpack_endian, 'U', 4294901244, b'\xfc\xfd\xfe\xff', pwny.Target.Endian.little
    yield check_short_form_unpack_endian, 'U', 4294901244, b'\xff\xfe\xfd\xfc', pwny.Target.Endian.big


def check_short_form_pack(f, num, bytestr):
    assert getattr(pwny, f)(num) == bytestr


def check_short_form_pack_endian(f, num, bytestr, endian):
    assert getattr(pwny, f)(num, endian=endian) == bytestr


def check_short_form_unpack(f, num, bytestr):
    assert getattr(pwny, f)(bytestr) == num


def check_short_form_unpack_endian(f, num, bytestr, endian):
    assert getattr(pwny, f)(bytestr, endian=endian) == num



from nose.tools import raises
import pwny


SOURCE_ASM = 'mov al, [0xced]'
RESULT_BIN_32 = b'\xa0\xed\x0c\x00\x00'
RESULT_BIN_64 = b'\x8a\x04%\xed\x0c\x00\x00'


SOURCE_DISASM = b'\x5f'
RESULT_DISASM_32 = ['pop edi']
RESULT_DISASM_64 = ['pop rdi']


target_x86_32 = pwny.Target('x86', 32)
target_x86_64 = pwny.Target('x86', 64)
target_arm_32 = pwny.Target('arm', 32)


def test_asm_with_default_target():
    # Note: set up sets default arch to x86 32bit
    assert pwny.asm(SOURCE_ASM) == RESULT_BIN_32


def test_asm_with_target_x86_32():
    assert pwny.asm(SOURCE_ASM, target=target_x86_32) == RESULT_BIN_32


def test_asm_with_target_x86_64():
    assert pwny.asm(SOURCE_ASM, target=target_x86_64) == RESULT_BIN_64


@raises(SyntaxError)
def test_asm_syntax_error():
    pwny.asm('mov ced, 3')


@raises(NotImplementedError)
def test_asm_unsupported_target():
    pwny.asm(SOURCE_ASM, target=target_arm_32)


def test_disasm_with_default_target():
    # Note: set up sets default arch to x86 32bit
    assert pwny.disasm(SOURCE_DISASM) == RESULT_DISASM_32


def test_disasm_with_target_x86_32():
    assert pwny.disasm(SOURCE_DISASM, target=target_x86_32) == RESULT_DISASM_32


def test_disasm_with_target_x86_64():
    print(RESULT_DISASM_64, pwny.disasm(SOURCE_DISASM, target=target_x86_64))
    assert pwny.disasm(SOURCE_DISASM, target=target_x86_64) == RESULT_DISASM_64


@raises(NotImplementedError)
def test_disasm_unsupported_target():
    pwny.disasm(SOURCE_DISASM, target=target_arm_32)
